import os
import argparse
import datetime
import re

import pandas as pd
import numpy as np

from . import hgmd
from . import visualize as vis
from . import quads
import sys
import multiprocessing
import time
import math
import matplotlib.pyplot as plt
import random
import scipy.io
import csv
#from docs.source import conf


def init_parser(parser):
    """Initialize parser args."""
    parser.add_argument(
        'marker', type=str,
        help=("Marker file input")
    )
    parser.add_argument(
        'vis', type=str,
        help=("vis file input")
    )
    parser.add_argument(
        'cluster', type=str,
        help=("Cluster file input")
    )
    parser.add_argument(
        '-g', nargs='?', default=None,
        help="Optional Gene list"
    )
    parser.add_argument(
        'output_path', type=str,
        help="the output directory where output files should go"
    )
    parser.add_argument(
        '-C', nargs='?', default=None,
        help="Num of cores avail for parallelization"
    )
    parser.add_argument(
        '-X', nargs='?', default=None,
        help="X argument for XL-mHG"
    )
    parser.add_argument(
        '-L', nargs='?', default=None,
        help="L argument for XL-mHG"
    )
    parser.add_argument(
        '-Abbrev', nargs='?',default=[],
        help="Choose between abbreviated or full 3-gene computation"
    )
    parser.add_argument(
        '-K', nargs='?',default=None,
        help="K-gene combinations to include"
    )
    parser.add_argument(
        '-Down', nargs='?',default=False,
        help="Downsample"
    )
    parser.add_argument(
        '-Trim', nargs='?',default=2000,
        help="Trim output files"
    )
    parser.add_argument(
        '-Count', nargs='?',default=False,
        help="Set to True when count data is being used, for visualizations."
    )
    parser.add_argument(
        '-tenx', nargs='?',default=False,
        help="Set to True when count data is being used, for visualizations."
    )
    parser.add_argument(
        '-online', nargs='?',default=False,
        help="Set to True for online version."
    )
    parser.add_argument(
        '-skipvis', nargs='?',default=False,
        help="Set to True to skip visualizations."
    )
    return parser


def read_data(cls_path, tsne_path, marker_path, gene_path, D, tenx, online,skipvis):
    """
    Reads in cluster series, tsne data, marker expression without complements
    at given paths.
    """
    cls_ser = pd.read_csv(
        cls_path, sep='\t', index_col=0, names=['cell', 'cluster'], squeeze=True
    )
    if ',' in cls_ser.index[1]:
        cls_ser = pd.read_csv(
        cls_path, sep=',', index_col=0, names=['cell', 'cluster'], squeeze=True )

    if skipvis == 1:
        tsne = None
        pass
    else:
        tsne = pd.read_csv(
            tsne_path, sep='\t', index_col=0, names=['cell', 'tSNE_1', 'tSNE_2']
        )
        if np.isnan(tsne['tSNE_1'][0]):
            tsne = pd.read_csv(
            tsne_path, sep=',', index_col=0, names=['cell', 'tSNE_1', 'tSNE_2'] )

    start_= time.time()
    tenx = int(tenx)
    if tenx == 1:
        print('Loading 10X matrix')
        mat = scipy.io.mmread(marker_path+"matrix.mtx")
        features_path = marker_path + "genes.tsv"
        gene_names = [row[1] for row in csv.reader(open(features_path), delimiter="\t")]
        barcodes_path = marker_path + "barcodes.tsv"
        barcodes = [row[0] for row in csv.reader(open(barcodes_path), delimiter="\t")]
        #construct pandas dataframe w/ the pieces (gene names, barcodes, counts in sparse form)
        matrix = pd.DataFrame(index = gene_names, columns = barcodes )
        now = time.time()
        print('assembling expression matrix')
        for i,j,v in zip(mat.row, mat.col, mat.data):
            matrix.iat[i,j] = v
        matrix.fillna(0,inplace=True)
        noww = time.time()
        print(str(noww-now) + ' seconds')
        print('size: ' + str(matrix.shape))
        no_complement_marker_exp = matrix
        no_complement_marker_exp.rename_axis('cell',axis=1,inplace=True)
    else:

        #Should allow either tab OR comma delimited formats
        try:
            no_complement_marker_exp = pd.read_csv(
                marker_path,sep='\t', index_col=0
                ).rename_axis('cell',axis=1)
            if len(no_complement_marker_exp.columns) == 0:
                raise Exception
            elif len(no_complement_marker_exp.index) == 0:
                raise Exception
            else:
                pass
        except:
            no_complement_marker_exp = pd.read_csv(
                marker_path,sep=',', index_col=0
                ).rename_axis('cell',axis=1)
           
    if no_complement_marker_exp.shape[1] == cls_ser.shape[0]:
        pass
    else:
        for index,row in cls_ser.iteritems():
            if str(index) in list(no_complement_marker_exp):
                continue
            else:
                cls_ser.drop(labels=index,inplace=True)
    
    #gene list filtering
    no_complement_marker_exp = no_complement_marker_exp.loc[~no_complement_marker_exp.index.duplicated(keep='first')]
    no_complement_marker_exp = np.transpose(no_complement_marker_exp)
    no_complement_marker_exp.columns = [x.upper() for x in no_complement_marker_exp.columns]
    no_complement_marker_exp = no_complement_marker_exp.loc[~no_complement_marker_exp.index.duplicated(keep='first')]
    #gene filtering
    #-------------#
    if gene_path is None:
        pass
    else:

        #read the genes
        #Compatible with single line comma list OR one per line no commas OR mix of both
        master_gene_list = []
        with open(gene_path, "r") as genes:
            lines = genes.readlines()
            if len(lines) == 1:
                with open(gene_path, "r") as genes:
                    init_read = genes.read().splitlines()
                    master_str = str.upper(init_read[0])
                    master_gene_list = master_str.split(",")
            else:
                for i, line in enumerate(lines):
                    if '\n' in line:
                        master_gene_list.append(line[:-1])
                    else:
                        master_gene_list.append(line)
                for item in master_gene_list[:]:
                    if ',' in item:
                        new_split = item.split(",")
                        master_gene_list.remove(item)
                        for ele in new_split:
                            master_gene_list.append(str.upper(ele))

    
        new_no_comp_mark_exp = pd.DataFrame()
        master_gene_list = [y.upper() for y in master_gene_list]
        for gene in master_gene_list:
            try:
                new_no_comp_mark_exp[gene] = no_complement_marker_exp[gene]
            except:
                pass

        no_complement_marker_exp = new_no_comp_mark_exp
        '''
        for column_name in no_complement_marker_exp.columns:
            if str.upper(column_name) in master_gene_list:
                pass
            else:
                try:
                    no_complement_marker_exp.drop(column_name, axis=1, inplace=True)
                except:
                    pass
        '''
    #-------------#

    #downsampling
    #-------------#
    
    if D is False:
        pass
    else:
        # get number of genes to set downsample threshold
        gene_numb = len(no_complement_marker_exp.columns)
        #print(gene_numb)

        if gene_numb > 3000:
            if int(D) < int(2500):
                pass
            else:
                D = int(2500)
        #total number of cells input
        N = len(cls_ser)
        #print(N)
        #downsample target
        M = int(D)
        if N <= M:
            return (cls_ser, tsne, no_complement_marker_exp, gene_path)
        clusters = sorted(cls_ser.unique())
        counts = { x : 0 for x in clusters}
        for clus in cls_ser:
            counts[clus] = counts[clus]+1
        #at this point counts has values for # cells in cls
        #dict goes like ->{ cluster:#cells }
        take_nums = {x : 0 for x in clusters}
        for clstr in take_nums:
            take_nums[clstr] = math.ceil(counts[clstr]*(M/N))
        summ = 0
        for key in take_nums:
            summ = summ + take_nums[key]
        #print('Downsampled cell num ' + str(summ))
        counts= { x : 0 for x in clusters}
        new_cls_ser = cls_ser.copy(deep=True)
        keep_first = 0
        for index,value in new_cls_ser.iteritems():
            keep_first = keep_first + 1
            if keep_first ==1:
                counts[value] = counts[value]+1
                continue
            new_cls_ser.drop(index,inplace=True)
        cls_ser.drop(cls_ser.index[0],inplace=True)
        #Now new_cls_ser has all removed except first item, which we can keep
        for num in range(N-1):
            init_rand_num = random.randint(0,N-num-1-1)
            if counts[cls_ser[init_rand_num]] >= take_nums[cls_ser[init_rand_num]]:
                cls_ser.drop(cls_ser.index[init_rand_num], inplace=True)
                continue
            new_cls_ser = new_cls_ser.append(pd.Series([cls_ser[init_rand_num]], index=[cls_ser.index[init_rand_num]]))
            counts[cls_ser[init_rand_num]] = counts[cls_ser[init_rand_num]]+1
            cls_ser.drop(cls_ser.index[init_rand_num], inplace=True)
            
        new_cls_ser.rename_axis('cell',inplace=True)
        new_cls_ser.rename('cluster', inplace=True)
        return(new_cls_ser,tsne,no_complement_marker_exp,gene_path)


    #-------------#

            
    return (cls_ser, tsne, no_complement_marker_exp, gene_path)

def process(cls,X,L,plot_pages,cls_ser,tsne,marker_exp,gene_file,csv_path,vis_path,pickle_path,cluster_number,K,abbrev,cluster_overall,Trim,count_data,skipvis):
    #for cls in clusters:
    # To understand the flow of this section, read the print statements.
    heur_limit = min(50,len(marker_exp.columns))
    start_cls_time = time.time()
    print('########\n# Processing cluster ' + str(cls) + '...\n########')
    print(str(K) + ' gene combinations')
    if len(abbrev) == 0:
        pass
    else:
        print('Heuristic limit set to: ' + str(heur_limit))
    print('Running t test on singletons...')
    try:
        t_test = hgmd.batch_stats(marker_exp, cls_ser, cls)
    except Exception as err:
        print('t-test error')
        print(err)
    print('Calculating fold change')
    try:
        fc_test = hgmd.batch_fold_change(marker_exp, cls_ser, cls)
    except Exception as err:
        print('fold-change error')
        print(err)
    print('Running XL-mHG on singletons...')
    try:
        xlmhg = hgmd.batch_xlmhg(marker_exp, cls_ser, cls, X=X, L=L)
    except Exception as err:
        print('XLMHG error')
        print(err)
    try:
        q_val = hgmd.batch_q(xlmhg)
    except Exception as err:
        print('q-val error')
        print(err)
    # We need to slide the cutoff indices before using them,
    # to be sure they can be used in the real world. See hgmd.mhg_slide()
    try:
        cutoff_value = hgmd.mhg_cutoff_value(
            marker_exp, xlmhg[['gene_1', 'mHG_cutoff']]
            )
        xlmhg = xlmhg[['gene_1', 'mHG_stat', 'mHG_pval']].merge(
            hgmd.mhg_slide(marker_exp, cutoff_value), on='gene_1'
            )
        # Update cutoff_value after sliding
        cutoff_value = pd.Series(
            xlmhg['cutoff_val'].values, index=xlmhg['gene_1']
            )
        xlmhg = xlmhg\
          .sort_values(by='mHG_stat', ascending=True)
    except Exception as err:
        print('error in sliding values')
        print(err)
    print('Creating discrete expression matrix...')
    try:
        discrete_exp = hgmd.discrete_exp(marker_exp, cutoff_value, abbrev, xlmhg)
    except Exception as err:
        print('discrete matrix construction failed')
        print(err)

    '''
    #For checking the sliding issue
    count = 0
    print(discrete_exp['Reg4'].sort_values(ascending=False).head(667))
    #time.sleep(100000)
    print(cls_ser)
    for index in discrete_exp['Reg4'].sort_values(ascending=False).head(667).iteritems():
        for index2 in cls_ser.iteritems():
            if index[0] == index2[0]:
                if index2[1] == 0:
                    count = count +1
    print(count)
    print(discrete_exp['Reg4'].sort_values(ascending=False).head(70))
    print(cls_ser.to_string())
    #print(marker_exp['1600029D21Rik'].sort_values(ascending=False).head(160))
    #time.sleep(100000)
    '''
    if skipvis == 0:
        discrete_exp_full = discrete_exp.copy()
    print('Finding simple true positives/negatives for singletons...')
    #Gives us the singleton TP/TNs for COI and for rest of clusters
    #COI is just a DF, rest of clusters are a dict of DFs
    (sing_tp_tn, other_sing_tp_tn) = hgmd.tp_tn(discrete_exp, cls_ser, cls, cluster_overall)
    ### Take out any genes with a true positive less than 15% from the expression matrix ###
    discrete_exp = discrete_exp[ sing_tp_tn['gene_1'][ sing_tp_tn['TP'] > .15 ].values ]
    ########################################################################################
    ###########
    #OLD HEURISTICS
    #abb = '3'
    #if abb in abbrev:
    #    print('Heuristic Abbreviation initiated for ' + str(abbrev) )
    #    count = 0
    #    trips_list=[]
    #    for index,row in xlmhg.iterrows():
    #        if sing_tp_tn.set_index('gene_1').at[row['gene_1'],'TP'] <= .15:
    #            continue
    #        ##THREE GENE VARIABLE HERE##
    #        if count == trips_heur:
    #            break
    #        else:
    #            trips_list.append(row['gene_1'])
    #            count = count + 1
    #else:
    #    trips_list = None
    ############
    print('Finding pair expression matrix...')
    (
        gene_map, in_cls_count, pop_count,
        in_cls_product, total_product, upper_tri_indices,
        cluster_exp_matrices, cls_counts
    ) = hgmd.pair_product(discrete_exp, cls_ser, cls, cluster_number,cluster_overall)
    if K >= 4:
        print('')
        print('Starting quads')
        print('')
        quads_in_cls, quads_total, quads_indices, odd_gene_mapped, even_gene_mapped = quads.combination_product(discrete_exp,cls_ser,cls,xlmhg)
        print('')
        print('')
        print('')
        print('')
        print('HG TEST ON QUADS')
        quads_fin = quads.quads_hg(gene_map,in_cls_count,pop_count,quads_in_cls,quads_total,quads_indices,odd_gene_mapped,even_gene_mapped)

    if K == 3:
        start_trips = time.time()
        print('Finding Trips expression matrix...')
        try:
            trips_in_cls,trips_total,trips_indices,gene_1_mapped,gene_2_mapped,gene_3_mapped = hgmd.combination_product(discrete_exp,cls_ser,cls,abbrev,heur_limit)
        except Exception as err:
            print('error in 3-gene matrix construction')
            print(err)
        end_trips = time.time()
        print(str(end_trips-start_trips) + ' seconds')

    HG_start = time.time()
    
    print('Running hypergeometric test on pairs...')
    pair, revised_indices = hgmd.pair_hg(
        gene_map, in_cls_count, pop_count,
        in_cls_product, total_product, upper_tri_indices, abbrev, heur_limit
    )
    
    pair_q = hgmd.pairs_q(pair)
    
    HG_end = time.time()
    print(str(HG_end-HG_start) + ' seconds')
    pair_out_initial = pair\
    .sort_values(by='HG_pval', ascending=True)
    
    pair_out_initial['rank'] = pair_out_initial.reset_index().index + 1
    #spair_out_print = pair_out_initial.head(Trim)
    pair_out_initial.to_csv(
    csv_path + '/cluster_' + str(cls) + '_pair_full_unranked.csv'
    )
    if K == 3:
        HG_start = time.time()
        print('Running hypergeometric test & TP/TN on trips...')
        trips = hgmd.trips_hg(
            gene_map,in_cls_count,pop_count,
            trips_in_cls,trips_total,trips_indices,
            gene_1_mapped,gene_2_mapped,gene_3_mapped,abbrev,heur_limit
            )
        #print(trips)
        HG_end = time.time()
        print(str(HG_end-HG_start) + ' seconds')


    # Pair TP/TN FOR THIS CLUSTER
    print('Finding simple true positives/negatives for pairs...')
    try:
        pair_tp_tn = hgmd.pair_tp_tn(
            gene_map, in_cls_count, pop_count,
            in_cls_product, total_product, upper_tri_indices, abbrev, revised_indices
        )
    except Exception as err:
        print(err)
    #accumulates pair TP/TN vals for all other clusters
    ##NEW
    try:
        other_pair_tp_tn = {}
        for key in cluster_exp_matrices:
            new_pair_tp_tn = hgmd.pair_tp_tn(
                gene_map, cls_counts[key], pop_count,
                cluster_exp_matrices[key], total_product, upper_tri_indices,
                abbrev, revised_indices
            )
            other_pair_tp_tn[key] = new_pair_tp_tn
            other_pair_tp_tn[key].set_index(['gene_1','gene_2'],inplace=True)
    except:
        pass
    try:
        pair = pair\
          .merge(pair_tp_tn, on=['gene_1', 'gene_2'], how='left')\
          .merge(pair_q, on=['gene_1','gene_2'], how='left')
    except:
        pass
        
    #Supplementary figure 2 code#
    '''
    for cl in other_pair_tp_tn:
        print(cl)
        print(other_pair_tp_tn[cl].loc[('LY6D', 'CD3G_negation')])
    for cl in other_sing_tp_tn:
        print(cl)
        print(other_sing_tp_tn[cl].loc[('LY6D')])
    time.sleep(10000)
    '''
    try:
        pair_tp_tn.set_index(['gene_1','gene_2'],inplace=True)
    except:
        pass
#    sing_tp_tn.set_index(['gene_1'], inplace=True)
    rank_start = time.time()
    print('Finding NEW Rank')
    try:
        ranked_pair,histogram = hgmd.ranker(pair,xlmhg,sing_tp_tn,other_sing_tp_tn,other_pair_tp_tn,cls_counts,in_cls_count,pop_count)
    except Exception as err:
        print(err)
    rank_end = time.time()
    print(str(rank_end - rank_start) + ' seconds')
    # Save TP/TN values to be used for non-cluster-specific things
    print('Pickling data for later...')
    try:
        sing_tp_tn.to_pickle(pickle_path + 'sing_tp_tn_' + str(cls))
        pair_tp_tn.to_pickle(pickle_path + 'pair_tp_tn_' + str(cls))
    except:
        print('pickling failed')
    #trips_tp_tn.to_pickle(pickle_path + 'trips_tp_tn' + str(cls))
    print('Exporting cluster ' + str(cls) + ' output to CSV...')
    try:
        sing_output = xlmhg\
          .merge(t_test, on='gene_1')\
          .merge(fc_test, on='gene_1')\
          .merge(sing_tp_tn, on='gene_1')\
          .merge(q_val, on='gene_1')\
          .set_index('gene_1')\
          .sort_values(by='mHG_stat', ascending=True)
    except Exception as err:
        print(err)
        sing_output = xlmhg.sort_values(by='mHG_stat',ascending=True)
    sing_output.sort_values(by='gene_1',ascending=True).to_csv(
        csv_path + '/cluster_' + str(cls) + '_singleton_full_unranked.csv'
    )

    sing_output = sing_output.loc[sing_output['TP'] >= .15]

    for index, row in sing_output.iterrows():
        if re.compile(".*_negation$").match(index):
            if row['TN'] >= .5:
                pass
            else:
                sing_output.drop(labels=index,axis=0,inplace=True)


    # make a copy of sing_output
    # run the stuff below on both the original (fine) and the copy (no negations)
    sing_output_positives = sing_output.copy(deep=True)
    for index, row in sing_output_positives.iterrows():
        if re.compile(".*_negation$").match(index):
            sing_output_positives.drop(labels=index,axis=0,inplace=True)
    sing_output_positives['hgrank'] = sing_output_positives.reset_index().index + 1
    sing_output_positives.sort_values(by='Log2FoldChangeAbs', ascending=False, inplace=True)
    sing_output_positives['fcrank'] = sing_output_positives.reset_index().index + 1
    sing_output_positives['finrank'] = sing_output_positives[['hgrank', 'fcrank']].mean(axis=1)
    sing_output_positives.sort_values(by='finrank',ascending=True,inplace=True)
    sing_output_positives['rank'] = sing_output_positives.reset_index().index + 1
    sing_output_positives.drop('finrank',axis=1, inplace=True)
    count = 1
    for index,row in sing_output_positives.iterrows():
        if count == 100:
            break
        sing_output_positives.at[index,'Plot'] = 1
        count = count + 1
    sing_output_positives.to_csv(
        csv_path + '/cluster_' + str(cls) + '_singleton_positive_markers_ranked.csv'
    )

                
    sing_output['hgrank'] = sing_output.reset_index().index + 1
    sing_output.sort_values(by='Log2FoldChangeAbs', ascending=False, inplace=True)
    sing_output['fcrank'] = sing_output.reset_index().index + 1
    sing_output['finrank'] = sing_output[['hgrank', 'fcrank']].mean(axis=1)
    sing_output.sort_values(by='finrank',ascending=True,inplace=True)
    sing_output['rank'] = sing_output.reset_index().index + 1
    sing_output.drop('finrank',axis=1, inplace=True)
    count = 1
    for index,row in sing_output.iterrows():
        if count == 100:
            break
        sing_output.at[index,'Plot'] = 1
        count = count + 1
        
    sing_output.to_csv(
        csv_path + '/cluster_' + str(cls) + '_singleton_all_ranked.csv'
    )
    sing_stripped = sing_output[
        ['mHG_stat', 'TP', 'TN']
    ].reset_index().rename(index=str, columns={'gene_1': 'gene_1'})
    
    try:
        ranked_print = ranked_pair.head(Trim)
        ranked_print.to_csv(
            csv_path + '/cluster_' + str(cls) + '_pair_final_ranking.csv'
        )
    except:
        print('pair file not generated, no pairs available')
        ranked_pair = pd.DataFrame(data=0,index=[1,2,3],columns={'TP','TN','Plot','gene_1','gene_2','rank'})
    #Add trips data pages
    #does not currently do new rank scheme
    if K == 3:
        trips_output = trips
        #  .sort_values(by='HG_stat', ascending=True)
          #print(trips_output)
        trips_output['rank'] = trips_output.reset_index().index + 1
        trips_print = trips_output.head(Trim)
        trips_print.to_csv(
            csv_path + '/cluster_' + str(cls) + '_trips.csv'
            )
    else:
        trips_output = int(1)
    if K >= 4:
        quads_final = quads_fin\
          .sort_values(by='HG_stat', ascending=True)
        quads_final['rank'] = quads_final.reset_index().index + 1
        quads_print = quads_final.head(Trim)
        quads_print.to_csv(
            csv_path + '/cluster_' + str(cls) + '_quads.csv'
            )
    else:
        quads_final = int(1)
    #plt.bar(list(histogram.keys()), histogram.values(), color='b')
    #plt.savefig(vis_path + '/cluster_' + str(cls) + '_pair_histogram')

    #if cls == fincls:
    # cls = 0
    if skipvis == 1:
        print('Skipping plots...')
        pass
    else:
        print('Drawing plots...')
        vis.make_plots(
            pair=ranked_pair,
            sing=sing_output,
            sing_tp_tn=sing_tp_tn,
            xlmhg=xlmhg,
            trips=trips_output,
            quads_fin=quads_final,
            tsne=tsne,
            discrete_exp=discrete_exp_full,
            marker_exp=marker_exp,
            plot_pages=plot_pages,
            combined_path=vis_path + '/cluster_' + str(cls) + '_pairs_as_singletons',
            sing_combined_path=vis_path + '/cluster_' +
            str(cls) + '_singleton',
            discrete_path=vis_path + '/cluster_' + str(cls) + '_discrete_pairs',
            tptn_path=vis_path + 'cluster_' + str(cls) + 'pair_TP_TN',
            trips_path=vis_path + 'cluster_' + str(cls) + '_discrete_trios',
            quads_path=vis_path + 'cluster_' + str(cls) + '_discrete_quads',
            sing_tptn_path=vis_path + 'cluster_' + str(cls) + '_singleton_TP_TN',
            count_data = count_data
            )
    end_cls_time=time.time()
    print(str(end_cls_time - start_cls_time) + ' seconds')
    #time.sleep(10000)


def main():
    """Hypergeometric marker detection. Finds markers identifying a cluster.

    Reads in data from single-cell RNA sequencing. Data is in the form of 3
    CSVs: gene expression data by gene by cell, 2-D tSNE data by cell, and the
    clusters of interest by cell. Creates a list of genes and a list of gene
    pairs (including complements), ranked by hypergeometric and t-test
    significance. The highest ranked marker genes generally best identify the
    cluster of interest. Saves these lists to CSV and creates gene expression
    visualizations.
    """
    # TODO: more precise description

    #ten_x = input("10X data? [y/n]")
    
    start_dt = datetime.datetime.now()
    start_time = time.time()
    print("Started on " + str(start_dt.isoformat()))
    
    args = init_parser(argparse.ArgumentParser(
        description=("Hypergeometric marker detection. Finds markers identifying a cluster. Documentation available at https://hgmd.readthedocs.io/en/latest/index.html")
    )).parse_args()
    
    output_path = args.output_path
    C = args.C
    K = args.K
    Abbrev = args.Abbrev
    Down = args.Down
    X = args.X
    L = args.L
    marker_file = args.marker
    tsne_file = args.vis
    cluster_file = args.cluster
    gene_file = args.g
    Trim = args.Trim
    count_data = args.Count
    tenx = args.tenx
    online = args.online
    skipvis = args.skipvis
    plot_pages = 30  # number of genes to plot (starting with highest ranked)

    # TODO: gene pairs with expression ratio within the cluster of interest
    # under [min_exp_ratio] were ignored in hypergeometric testing. This
    # functionality is currently unimplemented.
    # min_exp_ratio = 0.4

    csv_path = output_path + 'data/'
    vis_path = output_path + 'vis/'
    pickle_path = output_path + '_pickles/'
    try:
        os.makedirs(csv_path)
    except:
        os.system('rm -r ' + csv_path)
        os.makedirs(csv_path)

    try:
        os.makedirs(vis_path)
    except:
        os.system('rm -r ' + vis_path)
        os.makedirs(vis_path)

    try:
        os.makedirs(pickle_path)
    except:
        os.system('rm -r ' + pickle_path)
        os.makedirs(pickle_path)

    if Trim is not None:
        Trim = int(Trim)
    else:
        Trim = int(2000)
    if C is not None:
        C = abs(int(C))
    else:
        C = 1
    if X is not None:
        try:
            X = float(X)
        except:
            raise Exception('X param must be a number between 0 and 1')
        if X > 1:
            X = int(1)
        elif X <= 0:
            X = int(0)
        else:
            X = float(X)
        print("Set X to " + str(X) + ".")
    if L is not None:
        L = int(L)
        print("Set L to " + str(L) + ".")
    if K is not None:
        K = int(K)
    else:
        K = 2
    if K > 4:
        K = 4
        print('Only supports up to 4-gene combinations currently, setting K to 4')
    if count_data is not None:
        if count_data == str(True):
            count_data = 1
            print('Count Data')
        elif count_data == 'yes':
            count_data = 1
            print('Count Data')
        else:
            count_data = int(0)
    else:
        count_data = int(0)
    if tenx is not None:
        if tenx == str(True):
            tenx = int(1)
        elif tenx == 'yes':
            tenx = int(1)
        else:
            tenx = int(0)
    else:
        tenx = int(0)
    if online is not None:
        if online == str(True):
            online = int(1)
        elif online == 'yes':
            online = int(1)
        else:
            online = int(0)
    else:
        online = int(0)
    if skipvis is not None:
        if skipvis == str(True):
            skipvis = int(1)
        elif skipvis == 'yes':
            skipvis = int(1)
        else:
            skipvis = int(0)
    else:
        skipvis = int(0)
    print("Reading data...")
    if gene_file is None:
        (cls_ser, tsne, no_complement_marker_exp, gene_path) = read_data(
            cls_path=cluster_file,
            tsne_path=tsne_file,
            marker_path=marker_file,
            gene_path=None,
            D=Down,
            tenx=tenx,
            online=online,
            skipvis=skipvis)
    else:
        (cls_ser, tsne, no_complement_marker_exp, gene_path) = read_data(
            cls_path=cluster_file,
            tsne_path=tsne_file,
            marker_path=marker_file,
            gene_path=gene_file,
            D=Down,
            tenx=tenx,
            online=online,
            skipvis=skipvis)

    #throw out vals that show up in expression matrix but not in cluster assignments
    cls_ser_idx = set(cls_ser.index.values.tolist())
    no_complement_marker_exp.drop([ind for ind,row in no_complement_marker_exp.iterrows() if ind not in cls_ser_idx], inplace=True)

    #throw out gene rows that are duplicates and print out a message to user
    '''
    #throw out cls_ser vals not in marker_exp
    for index in cls_ser.index.values.tolist():
        if index in no_complement_marker_exp.columns:
            continue
        else:
            cls_ser.drop(index,inplace=True)
    '''

    print("Generating complement data...")
    marker_exp = hgmd.add_complements(no_complement_marker_exp)
    marker_exp.sort_values(by='cell',inplace=True)
    cls_ser.sort_index(inplace=True)

    # Process clusters sequentially
    clusters = cls_ser.unique()
    clusters.sort()
    cluster_overall=clusters.copy()

    #Only takes a certain number of clusters (cuts out smallest ones)
    if online == 1:
        max_clus_size = 15
        if len(clusters) <= max_clus_size:
            pass
        else:
            cls_helper = list(clusters.copy())
            cls_size_count = {}
            for item in cls_ser:
                if item in cls_size_count:
                    cls_size_count[item] = cls_size_count[item] + 1
                else:
                    cls_size_count[item] = 1
            for counted in cls_size_count:
                cls_size_count[counted] = cls_size_count[counted] / len(cls_ser)
            while len(cls_helper) > max_clus_size:
                lowest = 1
                place = 0
                for key in cls_size_count:
                    if cls_size_count[key] < lowest:
                        place = key
                        lowest = cls_size_count[key]
                cls_helper.remove(place)
                del cls_size_count[place]
            clusters = np.array(cls_helper)

            
    #Below could probably be optimized a little (new_clust not necessary),
    #cores is number of simultaneous threads you want to run, can be set at will
    cores = C
    cluster_number = len(clusters)
    # if core number is bigger than number of clusters, set it equal to number of clusters
    if cores > len(clusters):
        cores = len(clusters)
    if cores == 1:
        for cls in clusters:
            process(cls,X,L,plot_pages,cls_ser,tsne,marker_exp,gene_file,csv_path,vis_path,pickle_path,cluster_number,K,Abbrev,cluster_overall,Trim,count_data,skipvis)
    else:
        #below loops allow for splitting the job based on core choice
        group_num  = math.ceil((len(clusters) / cores ))
        for element in range(group_num):
            new_clusters = clusters[:cores]
            print(new_clusters)
            jobs = []
            #this loop spawns the workers and runs the code for each assigned.
            #workers assigned based on the new_clusters list which is the old clusters
            #split up based on core number e.g.
            #clusters = [1 2 3 4 5 6] & cores = 4 --> new_clusters = [1 2 3 4], new_clusters = [5 6]
            for cls in new_clusters:
                p = multiprocessing.Process(target=process,
                    args=(cls,X,L,plot_pages,cls_ser,tsne,marker_exp,gene_file,csv_path,vis_path,pickle_path,cluster_number,K,Abbrev,cluster_overall,Trim,count_data,skipvis))
                jobs.append(p)
                p.start()
            p.join()
            new_clusters = []
            clusters = clusters[cores:len(clusters)]

    end_time = time.time()

    

    # Add text file to keep track of everything
    end_dt = datetime.datetime.now()
    print("Ended on " + end_dt.isoformat())
    metadata = open(output_path + 'metadata.txt', 'w')
    metadata.write("Started: " + start_dt.isoformat())
    metadata.write("\nEnded: " + end_dt.isoformat())
    metadata.write("\nElapsed: " + str(end_dt - start_dt))
    #metadata.write("\nGenerated by COMET version " + conf.version)


    print('Took ' + str(end_time-start_time) + ' seconds')
    print('Which is ' + str( (end_time-start_time)/60 ) + ' minutes')

if __name__ == '__main__':
    main()



import pandas as pd
import numpy as np
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
from matplotlib import cm
from matplotlib.backends.backend_pdf import PdfPages
from itertools import repeat
import time
import os
import math
from adjustText import adjust_text

"""
Set of modularized visualization functions for COMET; producing graphs in PDFs.
"""

"""
Goals: Separate singleton and pair graphs for:

    - TP/TN with gene names

    - Side-by-side comparison between discrete and continuous expression

For singletons and pairs combined:

    - Side-by-side comparison of single gene discrete expression and resulting
      pair discrete expression.
"""


CMAP_CONTINUOUS = cm.get_cmap('nipy_spectral')
#CMAP_CONTINUOUS = cm.get_cmap('viridis')
CMAP_DISCRETE = cm.get_cmap('bwr')

def make_plots(
    pair, sing, sing_tp_tn, xlmhg, trips, quads_fin, tsne, discrete_exp, marker_exp, plot_pages,
    combined_path, sing_combined_path, discrete_path, tptn_path,trips_path,quads_path,sing_tptn_path,count_data
):
    """
    General function for all visualization generation.  Arguments should be
    self-explanatory.  See __main__.
    """

    # cutoff maps genes to their (absolute) cutoff. I.e. complements are mapped
    # to positive cutoffs.
    #print(discrete_exp)
    #print(marker_exp)
    cutoff = sing['cutoff_val'].abs()
    rank = sing['rank']
    TP_pair = pair['TP']
    TN_pair = pair['TN']
    TP_sing = sing['TP']
    TN_sing = sing['TN']
    xlmhg.set_index('gene_1',inplace=True)
    pair_sing_only = pair[pair['gene_2'].isnull()]
    sing_rank = pd.Series(
        pair_sing_only['rank'].values, index=pair_sing_only['gene_1']
    )

    p_short = pair[pair['Plot']==1]
    #p_short = pair.iloc[:plot_pages]
    s_short = sing[sing['Plot']==1]
    #s_short = sing.iloc[:plot_pages]

    neg_vals = 1

    for col in marker_exp:
        if min(marker_exp[col]) >= 0:
            pass
        else:
            neg_vals = 0

    
    if count_data == 1:
        marker_exp = marker_exp.applymap(lambda x: math.log(abs(x)+1,2))
        #marker_exp = marker_exp.applymap(lambda x: min(x,1500))
    if type(trips) == int:
        pass
    else:
        t_short = trips.iloc[:plot_pages]
    if type(quads_fin) == int:
        pass
    else:
        q_short=quads_fin.iloc[:plot_pages]
    try:
        vmt = np.vectorize(make_title,excluded=[6,7])
        d_plot_genes = zip(
            zip(
                vmt(
                    p_short['gene_1'], p_short['gene_2'],
                    p_short['rank'], p_short['gene_1'].map(cutoff),
                    p_short['TP'], p_short['TN'], sing_tp_tn,xlmhg
                    ), vmt(
                        p_short['gene_1'], np.nan,
                        p_short['gene_1'].map(rank),
                        p_short['gene_1'].map(cutoff),
                        p_short['gene_1'].map(TP_sing), p_short['gene_1'].map(TN_sing), sing_tp_tn,xlmhg
                        ), vmt(
                            p_short['gene_2'], np.nan,
                            p_short['gene_2'].map(rank),
                            p_short['gene_2'].map(cutoff),
                            p_short['gene_2'].map(TP_sing), p_short['gene_2'].map(TN_sing), sing_tp_tn,xlmhg
                            )
                ), p_short['gene_1'].values, p_short['gene_2'].values
            )
    except:
        print('No genes to plot')
    if type(trips) == int:
        pass
    else:
        try:
            vmt_2 = np.vectorize(make_trips_title)
            t_plot_genes = zip(
                zip(
                    vmt_2(
                        t_short['gene_1'], t_short['gene_2'], t_short['gene_3'],
                        t_short['rank'], t_short['gene_1'].map(cutoff)
                        ), vmt_2(
                            t_short['gene_1'], np.nan, np.nan,
                            t_short['gene_1'].map(rank),
                            t_short['gene_1'].map(cutoff)
                            ), vmt_2(
                                t_short['gene_2'], np.nan, np.nan,
                                t_short['gene_2'].map(rank),
                                t_short['gene_2'].map(cutoff)
                                ), vmt_2(
                                    t_short['gene_3'], np.nan, np.nan,
                                    t_short['gene_3'].map(rank),
                                    t_short['gene_3'].map(cutoff)
                                    )
                    ), t_short['gene_1'].values, t_short['gene_2'].values, t_short['gene_3'].values
                )
        except:
            print('No genes to plot')
    

    if type(quads_fin) == int:
        pass
    else:
        vmt_3 = np.vectorize(make_quads_title)
        q_plot_genes = zip(
            zip(
                vmt_3(
                    q_short['gene_1'], q_short['gene_2'], q_short['gene_3'], q_short['gene_4'],
                    q_short['rank'], q_short['gene_1'].map(cutoff)
                    ), vmt_3(
                        q_short['gene_1'], np.nan, np.nan, np.nan,
                        q_short['gene_1'].map(rank),
                        q_short['gene_1'].map(cutoff)
                        ), vmt_3(
                            q_short['gene_2'], np.nan, np.nan, np.nan,
                            q_short['gene_2'].map(rank),
                            q_short['gene_2'].map(cutoff)
                            ), vmt_3(
                                q_short['gene_3'], np.nan, np.nan, np.nan,
                                q_short['gene_3'].map(rank),
                                q_short['gene_3'].map(cutoff)
                                ), vmt_3(
                                q_short['gene_3'], np.nan, np.nan, np.nan,
                                q_short['gene_3'].map(rank),
                                q_short['gene_3'].map(cutoff)
                                )
                ), q_short['gene_1'].values, q_short['gene_2'].values, q_short['gene_3'].values, q_short['gene_4'].values)


    
    print("Drawing discrete plots for pairs...")
    try:
        make_discrete_plots(
            tsne, discrete_exp, d_plot_genes, discrete_path, 2, neg_vals
        )
    except Exception as err:
        print('No plots generated')
        print(err)
    if type(trips) == int:
        pass
    else:
        print("Drawing discrete plots for trips...")
        try:
            make_discrete_plots(
                tsne, discrete_exp, t_plot_genes, trips_path, 3, neg_vals
                )
        except Exception as err:
            print('no plots generated')
            print(err)
    
    if type(quads_fin) == int:
        pass
    else:
        print("Drawing discrete plots for quads...")
        try:
            make_discrete_plots(
                tsne, discrete_exp, q_plot_genes, quads_path, 4, neg_vals
                )
        except Exception as err:
            print('no plots generated')
            print(err)
    try:
        c_plot_genes = zip(
            zip(
                vmt(
                    p_short['gene_1'], np.nan,
                    p_short['gene_1'].map(rank),
                    p_short['gene_1'].map(cutoff),
                    p_short['gene_1'].map(TP_sing),
                    p_short['gene_1'].map(TN_sing),sing_tp_tn,xlmhg
                    ), vmt(
                        p_short['gene_2'], np.nan,
                        p_short['gene_2'].map(rank),
                        p_short['gene_2'].map(cutoff),
                        p_short['gene_2'].map(TP_sing),
                        p_short['gene_2'].map(TN_sing),sing_tp_tn,xlmhg
                        )
                ), p_short['gene_1'].values, p_short['gene_2'].values
            )
    except:
        print('no genes to plot')
    print("Drawing combined plots...")
    try:
        make_combined_plots(
            tsne, discrete_exp, marker_exp, c_plot_genes, combined_path, neg_vals
            )
    except Exception as err:
        print('no plots generated')
        print(err)
    c_s_plot_genes = zip(
        zip(
            vmt(
                s_short.index, np.nan,
                s_short['rank'], s_short['cutoff_val'],
                s_short['TP'], s_short['TN'],sing_tp_tn,xlmhg
            ), repeat(np.nan)
        ), s_short.index, repeat(np.nan)
    )
    print("Drawing singleton combined plots...")
    try:
        make_combined_plots(
            tsne, discrete_exp, marker_exp, c_s_plot_genes, sing_combined_path, neg_vals
            )
    except Exception as err:
        print('no plots generated')
        print(err)
    pair_tp_tn = pair[['gene_1', 'gene_2', 'TP', 'TN']]
    sing_tp_tn = sing[['TP', 'TN']]
    print("Drawing true positive/negative plots...")
    try:
        make_tp_tn_plot(
            zip(p_short['gene_1'], p_short['gene_2']),
            sing_tp_tn, pair_tp_tn, tptn_path, 0
            )
        make_tp_tn_plot(
            zip(s_short.index, repeat(np.nan)),
            sing_tp_tn, pair_tp_tn, sing_tptn_path, 1
            )
    except Exception as err:
        print('no plots generated')
        print(err)

def make_title(gene_1, gene_2, rank, cutoff_val, TP, TN, sing_tp_tn,xlmhg):
    """Makes a plot title for a gene or gene pair.

    Formatting: for pairs, 'rank $rank: $gene_1+$gene_2', and for singletons,
    'rank $rank: $gene_1 $cutoff_val'.  gene_2 should be None for singletons.

    :param genes: A DataFrame with columns 'gene_1', 'gene_2', 'rank',
        'cutoff_val'.

    :returns: A list of strings containing the titles, with indices
              corresponding to the input DataFrame.

    :rtype: string list
    """
    if np.isnan(rank):
        rank = 'Filtered'
        cutoff_val = xlmhg.at[gene_1,'cutoff_val']
        TP = sing_tp_tn.at[gene_1,'TP']
        TN = sing_tp_tn.at[gene_1,'TN']
    if pd.isna(gene_2):
        try:
            return ("rank %.0f: %s %.3f - TP:%.2f TN:%.2f" % (rank, gene_1, cutoff_val, TP, TN))
        except:
            return ("rank %s: %s %.3f - TP:%.2f TN:%.2f" % (rank, gene_1, cutoff_val, TP, TN))
    else:
        return ("rank %.0f: %s+%s - TP:%.2f TN:%.2f" % (rank, gene_1, gene_2, TP, TN))

def make_trips_title(gene_1,gene_2,gene_3,rank,cutoff_val):
    
    if pd.isna(gene_2):
        return ("rank %.0f: %s %.3f" % (rank, gene_1, cutoff_val))
    elif pd.isna(gene_3):
        return ("rank %.0f: %s+%s" % (rank, gene_1, gene_2))
    else:
        return ("rank %.0f: %s+%s+%s" % (rank, gene_1, gene_2, gene_3)) 

def make_quads_title(gene_1,gene_2,gene_3,gene_4,rank,cutoff_val):

    
    if pd.isna(gene_2):
        return ("rank %.0f: %s %.3f" % (rank, gene_1, cutoff_val))
    elif pd.isna(gene_3):
        return ("rank %.0f: %s+%s" % (rank, gene_1, gene_2))
    elif pd.isna(gene_4):
        return ("rank %.0f: %s+%s+%s" % (rank, gene_1, gene_2, gene_3))
    else:
        return ("rank %.0f: %s+%s+%s+%s" % (rank, gene_1, gene_2, gene_3, gene_4))


#Need another version of this for non-negations that doesnt do absolute value
def make_plot(neg_vals, ax, title, coords, cmap, draw_cbar=False):
    """
    Make a single graph on ax with given specs.  Plots only absolute values.
    """

    if neg_vals == 1:
    
        ax.set_title(title)#,fontdict={'fontsize':6})
        ax.set_xlabel('Vis_X')
        ax.set_ylabel('Vis_Y')
        sc = ax.scatter(
            x=coords[0],
            y=coords[1],
            c=abs(coords[2]),
            s=4,
            cmap=cmap
        )
        if draw_cbar:
            plt.colorbar(sc, ax=ax)
    else:
        ax.set_title(title)#,fontdict={'fontsize':6})
        ax.set_xlabel('Vis_X')
        ax.set_ylabel('Vis_Y')
        sc = ax.scatter(
            x=coords[0],
            y=coords[1],
            c=coords[2],
            s=4,
            cmap=cmap
        )
        if draw_cbar:
            plt.colorbar(sc, ax=ax)


def make_discrete_plots(tsne, discrete_exp, plot_genes, path,num,neg_vals):
    """Plots discrete gene expression of paired genes to PDF.

    For each gene pair listed in plot_genes, make three scatterplots.  First, a
    plot showing joint expression.  Then, two plots showing singleton
    expression for each genes.  If a single gene is passed, plot only its
    expression, and make two blank plots.  Save each gene/gene pair as a PDF
    page, then save to path.

    :param tsne: A DataFrame with 'cell', 'tSNE_1', and 'tSNE_2' columns.
    :param discrete_exp: A DataFrame whose rows are cell identifiers, columns
        are gene identifiers, and values are boolean values representing gene
        expression.
    :param plot_genes: A list of 3-tuples, where the first element of each
        tuple is another 3-tuple containing the three plot titles to be used.
        The other 2 elements are the gene names to be plotted.
    :param path: The path to which the PDF will be saved.

    :returns: Nothing.
    """
    def make_quads_discrete_page(neg_vals, fig, ax_triple, titles, gene_1, gene_2, gene_3,gene_4):
        """Make page with quads discrete plots given titles and genes."""
        coords_df = tsne.merge(discrete_exp[[gene_1, gene_2, gene_3,gene_4]], on='cell')
        coords_df['pair'] = coords_df[gene_1] * coords_df[gene_2]
        coords_df['trips'] = coords_df[gene_1] * coords_df[gene_2] * coords_df[gene_3]
        coords_df['quads_fin'] = coords_df[gene_1] * coords_df[gene_2] * coords_df[gene_3] * coords_df[gene_4]

        for (graph_index, z_label) in ((0, 'quads_fin'), (1, gene_1), (2, gene_2), (3, gene_3), (4, gene_4)):
            make_plot(
                neg_vals,
                ax=ax_triple[graph_index],
                title=titles[graph_index],
                coords=(
                    coords_df['tSNE_1'].values,
                    coords_df['tSNE_2'].values,
                    coords_df[z_label].values
                ),
                cmap=CMAP_DISCRETE
            )

    
    def make_trips_discrete_page(neg_vals, fig, ax_triple, titles, gene_1, gene_2, gene_3):
        """Make page with trips discrete plots given titles and genes."""
        coords_df = tsne.merge(discrete_exp[[gene_1, gene_2, gene_3]], on='cell')
        coords_df['pair'] = coords_df[gene_1] * coords_df[gene_2]
        coords_df['trips'] = coords_df[gene_1] * coords_df[gene_2] * coords_df[gene_3]

        for (graph_index, z_label) in ((0, 'trips'), (1, gene_1), (2, gene_2), (3, gene_3)):
            make_plot(
                neg_vals,
                ax=ax_triple[graph_index],
                title=titles[graph_index],
                coords=(
                    coords_df['tSNE_1'].values,
                    coords_df['tSNE_2'].values,
                    coords_df[z_label].values
                ),
                cmap=CMAP_DISCRETE
            )
        
    def make_pair_discrete_page(neg_vals, fig, ax_triple, titles, gene_1, gene_2):
        """Make page with three discrete plots given titles and genes."""
        coords_df = tsne.merge(discrete_exp[[gene_1, gene_2]], on='cell')
        coords_df['pair'] = coords_df[gene_1] * coords_df[gene_2]

        for (graph_index, z_label) in ((0, 'pair'), (1, gene_1), (2, gene_2)):
            make_plot(
                neg_vals,
                ax=ax_triple[graph_index],
                title=titles[graph_index],
                coords=(
                    coords_df['tSNE_1'].values,
                    coords_df['tSNE_2'].values,
                    coords_df[z_label].values
                ),
                cmap=CMAP_DISCRETE
            )

    def make_single_discrete_page(neg_vals, fig, ax_triple, title, gene):
        """Make page with one discrete plot given title and gene"""
        print(title)
        coords_df = tsne.merge(discrete_exp[[gene]], on='cell')
        make_plot(
            neg_vals,
            ax=ax_triple[0],
            title=title[0],
            coords=(
                coords_df['tSNE_1'].values,
                coords_df['tSNE_2'].values,
                coords_df[gene].values
            ),
            cmap=CMAP_DISCRETE
        )

    #with PdfPages(path) as pdf:
    count = 1
    try:
        os.makedirs(path)
    except:
        os.system('rm -r ' + path)
        os.makedirs(path)
    for plot_gene in plot_genes:
        # print(plot_gene)
        if num == 4:
            fig, ax_triple = plt.subplots(ncols=5, figsize=(15, 5))
            if pd.isnull(plot_gene[2]):
                make_single_discrete_page(
                    neg_vals,
                    fig=fig, ax_triple=ax_triple,
                    title=plot_gene[0],
                    gene=plot_gene[1]
                    )
            elif pd.isnull(plot_gene[3]):
                make_pair_discrete_page(
                    neg_vals,
                    fig=fig, ax_triple=ax_triple,
                    titles=plot_gene[0],
                    gene_1=plot_gene[1],
                    gene_2=plot_gene[2]
                    )
            elif pd.isnull(plot_gene[4]):
                make_trips_discrete_page(
                    neg_vals,
                    fig=fig, ax_triple=ax_triple,
                    titles=plot_gene[0],
                    gene_1=plot_gene[1],
                    gene_2=plot_gene[2],
                    gene_3=plot_gene[3]
                    )
            else:
                make_quads_discrete_page(
                    neg_vals,
                    fig=fig, ax_triple=ax_triple,
                    titles=plot_gene[0],
                    gene_1=plot_gene[1],
                    gene_2=plot_gene[2],
                    gene_3=plot_gene[3],
                    gene_4=plot_gene[4]
                    )
        elif num == 3:
            fig, ax_triple = plt.subplots(ncols=4, figsize=(15, 5))
            if pd.isnull(plot_gene[2]):
                make_single_discrete_page(
                    neg_vals,
                    fig=fig, ax_triple=ax_triple,
                    title=plot_gene[0],
                    gene=plot_gene[1]
                    )
            elif pd.isnull(plot_gene[3]):
                make_pair_discrete_page(
                    neg_vals,
                    fig=fig, ax_triple=ax_triple,
                    titles=plot_gene[0],
                    gene_1=plot_gene[1],
                    gene_2=plot_gene[2]
                    )
            else:
                make_trips_discrete_page(
                    neg_vals,
                    fig=fig, ax_triple=ax_triple,
                    titles=plot_gene[0],
                    gene_1=plot_gene[1],
                    gene_2=plot_gene[2],
                    gene_3=plot_gene[3]
                    )
        else:
            fig, ax_triple = plt.subplots(ncols=3, figsize=(15, 5))
            if pd.isnull(plot_gene[2]):
                make_single_discrete_page(
                    neg_vals,
                    fig=fig, ax_triple=ax_triple,
                    title=plot_gene[0],
                    gene=plot_gene[1]
                    )
            else:
                make_pair_discrete_page(
                    neg_vals,
                    fig=fig, ax_triple=ax_triple,
                    titles=plot_gene[0],
                    gene_1=plot_gene[1],
                    gene_2=plot_gene[2]
                    )
                
            #pdf.savefig(fig)
        if len(plot_gene) >= 4:
            plt.savefig( path + '/' + 'rank_' + str(count) + '.png')
        else:
            try:
                if ':' in str(plot_gene[0][0][:9]):
                    if ':' in str(plot_gene[0][0][:8]):
                        if ':' in str(plot_gene[0][0][:7]):
                            plt.savefig( path + '/' + str(plot_gene[0][0][:6]) + '.png')
                        else:
                            plt.savefig( path + '/' + str(plot_gene[0][0][:7]) + '.png')
                    else:
                        plt.savefig( path + '/' + str(plot_gene[0][0][:8]) + '.png')
                else:
                    plt.savefig( path + '/' + str(plot_gene[0][0][:9]) + '.png')
            except:
                plt.savefig( path + '/' + 'rank_' + str(count) + '.png')
        count=count+1
        plt.close(fig)

def make_combined_plots(tsne, discrete_exp, marker_exp, plot_genes, path, neg_vals):
    """Plots discrete alongside continuous expression to PDFF.

    For each gene/gene pair listed in plot_genes, make two scatterplots: a plot
    showing discrete expression, and a plot showing expression on a color
    spectrum.  For gene pairs, make these two plots separately for each gene.
    Save each gene/gene pair as a PDF page, then save to path.

    :param tsne: A DataFrame with 'cell', 'tSNE_1', and 'tSNE_2' columns.
    :param discrete_exp: A DataFrame whose rows are cell identifiers, columns
        are gene identifiers, and values are boolean values representing gene
        expression.
    :param marker_exp: A DataFrame whose rows are cell identifiers, columns are
        gene identifiers, and values are float values representing gene
        expression.
    :param plot_genes: A list of 3-tuples, where the first element of each
        tuple is another 2-tuple containing the two titles to be used..
        The other 2 elements are the gene names to be plotted.
    :param path: The path to which the PDF will be saved.

    :returns: Nothing.
    """
    def make_pair_combined_page(neg_vals, fig, axes, titles, gene_1, gene_2):
        """Make page with two pairs of plots for given genes."""
        disc_coords = tsne.merge(discrete_exp[[gene_1, gene_2]], on='cell')
        cont_coords = tsne.merge(marker_exp[[gene_1, gene_2]], on='cell')
        for (graph_index, z_label) in ((0, gene_1), (1, gene_2)):
            make_plot(
                neg_vals,
                ax=axes[graph_index][0], title=titles[graph_index],
                coords=(
                    disc_coords['tSNE_1'].values,
                    disc_coords['tSNE_2'].values,
                    disc_coords[z_label].values
                ),
                cmap=CMAP_DISCRETE
            )
            make_plot(
                neg_vals,
                ax=axes[graph_index][1], title=str(z_label),
                coords=(
                    cont_coords['tSNE_1'].values,
                    cont_coords['tSNE_2'].values,
                    cont_coords[z_label].values
                ),
                cmap=CMAP_CONTINUOUS, draw_cbar=True,
            )

    def make_single_combined_page(neg_vals, fig, title, axes, gene):
        """Make page with single pair of plot of given gene."""
        disc_coords = tsne.merge(discrete_exp[[gene]], on='cell')
        cont_coords = tsne.merge(marker_exp[[gene]], on='cell')
        make_plot(
            neg_vals,
            ax=axes[0], title=title,
            coords=(
                disc_coords['tSNE_1'].values,
                disc_coords['tSNE_2'].values,
                disc_coords[gene].values
            ),
            cmap=CMAP_DISCRETE
        )
        #if str(gene)[-8:] == 'NEGATION':

        #This is, for some reason, plotting the absolute value of the scores
        make_plot(
            neg_vals,
            ax=axes[1], title=str(gene),
            coords=(
                cont_coords['tSNE_1'].values,
                cont_coords['tSNE_2'].values,
                cont_coords[gene].values
            ),
            cmap=CMAP_CONTINUOUS, draw_cbar=True
        )

    #with PdfPages(path) as pdf:
    counter = 1
    try:
        os.makedirs(path)
    except:
        os.system('rm -r ' + path)
        os.makedirs(path)
    for plot_gene in plot_genes:
        if pd.isnull(plot_gene[2]):
            fig, axes = plt.subplots(ncols=2, figsize=(10, 5))
            make_single_combined_page(
                neg_vals, fig, plot_gene[0][0], axes, plot_gene[1]
            )
        else:
            fig, axes = plt.subplots(
                nrows=2, ncols=2, figsize=(10, 10)
            )
            make_pair_combined_page(
                neg_vals, fig, axes, plot_gene[0], plot_gene[1], plot_gene[2]
            )
        #pdf.savefig(fig)
        plt.savefig( path + '/' + 'rank_' + str(counter) + '.png')
        counter=counter+1
        plt.close(fig)


def make_tp_tn_plot(plot_genes, sing_tp_tn, pair_tp_tn, path,sing_pair):
    """Plots TP/TN rates of genes/pairs to PDF.

    For each gene/gene pair listed in plot_genes, plot their TP/TN rate on a
    scatterplot, labeling the point with the gene/gene pair name.  When done,
    output this scatterplot to PDF and save to path.

    :param plot_genes: An array whose elements are tuples representing gene
        pairs.  If the second element is empty, it represents a singleton.
    :param sing_tp_tn: A DataFrame with ''TP', and 'TN' columns, with gene
        indices
    :param pair_tp_tn: A DataFrame with 'gene_1', 'gene_2', 'TP', and 'TN'
        columns.
    :param path: The path to which the PDF will be saved.

    :returns: Nothing.
    """
    PADDING = 0.002

    fig = plt.figure(figsize=[15, 15])
    plt.xlabel("True positive")
    plt.ylabel("True negative")
    plt.title("True positive/negative")
    plt.axis([0.0, 1.1, 0.0, 1.1])

    def get_data(genes):
        if pd.isnull(genes[1]):
            title = genes[0]
            data_row = sing_tp_tn.loc[genes[0]]
        else:
            title = genes[0] + "+" + genes[1]
            data_row = pair_tp_tn[
                ((pair_tp_tn['gene_1'] == genes[0]) &
                 (pair_tp_tn['gene_2'] == genes[1]))
                | ((pair_tp_tn['gene_1'] == genes[1]) &
                   (pair_tp_tn['gene_2'] == genes[0]))
            ]
        return [title, data_row['TP'], data_row['TN']]

    coords_df = pd.DataFrame()
    data = list(map(get_data, list(plot_genes)))
    coords_df[['title', 'TP', 'TN']] = pd.DataFrame(
        data, columns=['title', 'TP', 'TN']
    )

    if sing_pair == 1:
        new_list = [['zero',0,0]]
        for index, row in coords_df.iterrows():
            new_list.append([row['title'],row['TP'],row['TN']])
        new_df = pd.DataFrame(new_list, columns = ['title','TP','TN'])    
        #print(new_df)
        concat_new_df = new_df.head(50)
        #concat_new_df.drop(0,inplace=True)
        plt.scatter(concat_new_df['TP'], concat_new_df['TN'], s=3,c='red')
        #print(coords_df)
        #time.sleep(10000)
        texts = []
        for x,y,s in zip(concat_new_df['TP'],concat_new_df['TN'],concat_new_df['title']):
            texts.append(plt.text(x, y, s, size=7))
        adjust_text(texts,arrowprops=dict(arrowstyle="->", color='blue', lw=0.5))


        
    else:
        new_list = [['zero',0,0]]
        for index, row in coords_df.iterrows():
            TP = row['TP'].tolist()[-1]
            TN = row['TN'].tolist()[-1]
            new_list.append([row['title'],TP,TN])
        new_df = pd.DataFrame(new_list, columns = ['title','TP','TN'])    
        #print(new_df)
        concat_new_df = new_df.head(50)
        #concat_new_df.drop(0,inplace=True)
        plt.scatter(concat_new_df['TP'], concat_new_df['TN'], s=3,c='red')
        #print(coords_df)
        #time.sleep(10000)
        texts = []
        for x,y,s in zip(concat_new_df['TP'],concat_new_df['TN'],concat_new_df['title']):
            texts.append(plt.text(x, y, s, size=7))
        adjust_text(texts,arrowprops=dict(arrowstyle="->", color='blue', lw=0.5))
        
    #for index, row in coords_df.iterrows():
    #    plt.annotate(row['title'], (row['TP'] + PADDING, row['TN'] + PADDING))
    fig.savefig(path + '.png')
    plt.close(fig)



##Copyright (c) 2012, Nicolo Fusi, University of Sheffield
##All rights reserved.

import scipy as sp
from scipy import interpolate


def estimate(pv, m=None, verbose=False, lowmem=False, pi0=None):
    """
    Estimates q-values from p-values
    Args
    =====
    m: number of tests. If not specified m = pv.size
    verbose: print verbose messages? (default False)
    lowmem: use memory-efficient in-place algorithm
    pi0: if None, it's estimated as suggested in Storey and Tibshirani, 2003.
         For most GWAS this is not necessary, since pi0 is extremely likely to be
         1
    """
    assert(pv.min() >= 0 and pv.max() <= 1), "p-values should be between 0 and 1"

    original_shape = pv.shape
    pv = pv.ravel()  # flattens the array in place, more efficient than flatten()

    if m is None:
        m = float(len(pv))
    else:
        # the user has supplied an m
        m *= 1.0

    # if the number of hypotheses is small, just set pi0 to 1
    if len(pv) < 100 and pi0 is None:
        pi0 = 1.0
    elif pi0 is not None:
        pi0 = pi0
    else:
        # evaluate pi0 for different lambdas
        pi0 = []
        lam = sp.arange(0, 0.90, 0.01)
        counts = sp.array([(pv > i).sum() for i in sp.arange(0, 0.9, 0.01)])
        for l in range(len(lam)):
            pi0.append(counts[l]/(m*(1-lam[l])))

        pi0 = sp.array(pi0)

        # fit natural cubic spline
        tck = interpolate.splrep(lam, pi0, k=3)
        pi0 = interpolate.splev(lam[-1], tck)
        if verbose:
            print("qvalues pi0=%.3f, estimated proportion of null features " % pi0)

        if pi0 > 1:
            if verbose:
                print("got pi0 > 1 (%.3f) while estimating qvalues, setting it to 1" % pi0)
            pi0 = 1.0

    assert(pi0 >= 0 and pi0 <= 1), "pi0 is not between 0 and 1: %f" % pi0

    if lowmem:
        # low memory version, only uses 1 pv and 1 qv matrices
        qv = sp.zeros((len(pv),))
        last_pv = pv.argmax()
        qv[last_pv] = (pi0*pv[last_pv]*m)/float(m)
        pv[last_pv] = -sp.inf
        prev_qv = last_pv
        for i in range(int(len(pv))-2, -1, -1):
            cur_max = pv.argmax()
            qv_i = (pi0*m*pv[cur_max]/float(i+1))
            pv[cur_max] = -sp.inf
            qv_i1 = prev_qv
            qv[cur_max] = min(qv_i, qv_i1)
            prev_qv = qv[cur_max]

    else:
        p_ordered = sp.argsort(pv)
        pv = pv[p_ordered]
        qv = pi0 * m/len(pv) * pv
        qv[-1] = min(qv[-1], 1.0)

        for i in range(len(pv)-2, -1, -1):
            qv[i] = min(pi0*m*pv[i]/(i+1.0), qv[i+1])

        # reorder qvalues
        qv_temp = qv.copy()
        qv = sp.zeros_like(qv)
        qv[p_ordered] = qv_temp

    # reshape qvalues
    qv = qv.reshape(original_shape)

    return qv



"""
Set of modularized components of COMET's HGMD testing.

For marker expression, float comparisions are fuzzy to 1e-3.  Marker expression
must therefore be normalized to a point where a difference of 0.001 is
insignificant.  I.e. 15.001 and 15.000 are treated as equivalent expression
values.
"""

import re
import pandas as pd
import numpy as np
import xlmhg as hg
import scipy.stats as ss
import time
import math
from sklearn.linear_model import LogisticRegression
from sklearn.metrics import log_loss
from . import qvalue

# TODO: idea, replace 'gene_1', 'gene_2', and 'gene' columns with
# indices/multi-indices

# Used for comparision of marker expression values.
FLOAT_PRECISION = 0.001

# the following function was taken from SciPy and modified to minimize memory consumption
# Copyright © 2001, 2002 Enthought, Inc.; Copyright © 2003-2019 SciPy Developers.
from collections import namedtuple
RanksumsResult = namedtuple('RanksumsResult', ('statistic', 'pvalue'))
def ranksums(x, y):
    x, y = map(np.asarray, (x, y))
    n1 = len(x)
    n2 = len(y)
    alldata = np.concatenate((x, y))

    alldata = np.ravel(np.asarray(alldata))
    sorter = np.argsort(alldata, kind='quicksort')

    inv = np.empty(sorter.size, dtype=np.intp)
    inv[sorter] = np.arange(sorter.size, dtype=np.intp)

    alldata = alldata[sorter]
    sorter = None
    alldata = np.r_[True, alldata[1:] != alldata[:-1]]
    dense = alldata.cumsum()[inv]

    count = np.r_[np.nonzero(alldata)[0], len(alldata)]
    ranked = .5 * (count[dense] + count[dense - 1] + 1)

    x = ranked[:n1]
    s = np.sum(x, axis=0)
    expected = n1 * (n1+n2+1) / 2.0
    z = (s - expected) / np.sqrt(n1*n2*(n1+n2+1)/12.0)
    prob = 2 * ss.distributions.norm.sf(abs(z))

    return RanksumsResult(z, prob)

def add_complements(marker_exp):
    """Adds columns representing gene complement to a gene expression matrix.

    Gene complements are represented simplistically: gene expression values for
    a given gene X are multiplied by -1 and become a new column, labeled X_negation.
    "High" expression values of X become "low" values of X_negation, and vice versa,
    where discrete expression corresponds to a "high" value, and discrete
    non-expression to a "low" value.

    marker_exp should have cell row labels, gene column labels, gene expression
    float values.

    :param marker_exp: gene expression DataFrame whose rows are cell
        identifiers, columns are gene identifiers, and values are float values
        representing gene expression.

    :returns: A DataFrame of same format as marker_exp, but with a new column
              added for each existing column label, representing the column
              label gene's complement.

    :rtype: pandas.DataFrame
    """
    negation = -marker_exp
    negation.columns = negation.columns + '_negation'
    return pd.concat((marker_exp,negation), axis=1)



def batch_xlmhg(marker_exp, c_list, coi, X=None, L=None):
    """Applies XL-mHG test to a gene expression matrix, gene by gene.

    Outputs a 3-column DataFrame representing statistical results of XL-mHG.

    :param marker_exp: A DataFrame whose rows are cell identifiers, columns are
        gene identifiers, and values are float values representing gene
        expression.
    :param c_list: A Series whose indices are cell identifiers, and whose
        values are the cluster which that cell is part of.
    :param coi: The cluster of interest.
    :param X: An integer to be used as argument to the XL-mHG test.
    :param L: An integer to be used as argument to the XL-mHG test.

    :returns: A matrix with arbitrary row indices, whose columns are the gene
              name, stat, cutoff, and pval outputs of the XL-mHG test; of
              float, int, and float type respectively.  Their names are 'gene',
              'mHG_stat', 'mHG_cutoff', and 'mHG_pval'.

    :rtype: pandas.DataFrame
    """
    # * 1 converts to integer
    mem_list = (c_list == coi) * 1
    #count_n = 0
    #count_2n = 0
    #Count the number of cells in the cluster, store into count_n
    count_n = np.sum(mem_list)
    '''
    for cell in mem_list:
        if cell == 1:
            count_n = count_n + 1
        else:
            continue
    '''
    #Twice the number of cells in the cluster
    #count_2n = count_n * 2
    #Set X and L params
    if X is None:
        X = np.int(.15*count_n)
    else:
        X = np.int(X*count_n)
    if L is None:
        if 2*count_n >= marker_exp.shape[0]:
            L = np.int(marker_exp.shape[0])
        else:
            L = np.int(2*count_n)
       #L = marker_exp.shape[0]
    print('X = ' + str(X))
    print('L = ' + str(L))
    print('Cluster size ' + str(count_n))
    xlmhg = marker_exp.apply(
        lambda col:
        hg.xlmhg_test(
            mem_list.reindex(
                col.sort_values(ascending=False).index
            ).values,
            X=X,
            L=L
        )
    )
    xlmhg_1 = marker_exp.apply(
        lambda col:
        hg.get_xlmhg_test_result(N=len(mem_list),
                                     indices=np.array(np.where(np.array(mem_list.reindex(col.sort_values(ascending=False).index).values==1))[0],dtype='int64').astype('uint16')
                                     ,X=X,L=L,pval_thresh=1e-12,escore_pval_thresh=1e-1,tol=1e-12).escore
        )
    output = pd.DataFrame()
    output['gene_1'] = xlmhg.index
    output[['mHG_stat', 'mHG_cutoff', 'mHG_pval']] = pd.DataFrame(
        xlmhg.values.tolist(),
        columns=['mHG_stat', 'mHG_cutoff', 'mHG_pval']
    )
    '''
    print(xlmhg_1)
    output['escore'] = np.array(xlmhg_1,dtype=float)
    output.fillna(0,inplace=True)
    print(output.sort_values(by='escore', ascending=False))
    time.sleep(1000)
    '''
    return output



def batch_q(xlmhg):
    #Takes in the xlmhg dataframe which includes gene names and pvalue assignments
    #Returns a list of q-value, corresponding to each gene
    try:
        q_vals = qvalue.estimate(np.array(xlmhg['mHG_pval']))
    except:
        print('error in qval calculation')
    q_val_out = pd.DataFrame(data=xlmhg['gene_1'].tolist(),columns=['gene_1'])
    q_val_out['q_value'] = q_vals

    return q_val_out

def pairs_q(pair):
    try:
        q_vals = qvalue.estimate(np.array(pair['HG_pval']))
        q_val_out_pair = pd.DataFrame(data=pair['gene_1'].tolist(),columns=['gene_1'])
        q_val_out_pair['gene_2'] = pair['gene_2']
        q_val_out_pair['q_value'] = q_vals
        return q_val_out_pair
    except:
        print('error in qval calculation')
        q_val_out_pair = pd.DataFrame(data=pair['gene_1'].tolist(),columns=['gene_1'])
        q_val_out_pair['gene_2'] = pair['gene_2']
        q_val_out_pair['q_value'] = 0
        return q_val_out_pair

def batch_stats_extended(marker_exp, c_list, coi):
    """Applies t test , wilcoxon test, and likelihood ratio test (Based on logistic regression)
    to a gene expression matrix, gene by gene. Also gives simple up versus down regulation test (difference between means).

    :param marker_exp: A DataFrame whose rows are cell identifiers, columns are
        gene identifiers, and values are float values representing gene
        expression.
    :param c_list: A Series whose indices are cell identifiers, and whose
        values are the cluster which that cell is part of.
    :param coi: The cluster of interest.

    :returns: A matrix with arbitary row indices whose columns are the gene, t
              statistic, then t p-value; the last two being of float type.
              Their names are 'gene', 't_stat' , 't_pval' , w_stat, w_pval , LRT_pval, up/down regulated

    :rtype: pandas.DataFrame
    """
    def LRT_LogReg(df):
        # Define model matrix and response
        X = np.matrix(df.drop('cluster', axis=1))
        y = df['cluster']
        # Train logistic regression with full model
        logreg1 = LogisticRegression(solver='lbfgs').fit(X,y)
        ll1 = -log_loss(y,logreg1.predict_proba(X),normalize=False)
        # Train logistic regression with null model (only intercept)
        logreg0 = LogisticRegression(solver='lbfgs').fit([[0]]*len(X) ,y)
        ll0 = -log_loss(y,logreg0.predict_proba(X),normalize=False)
        # Likelihood ratio test
        stat = 2*(ll1-ll0)
        pval = ss.chi2.sf(stat, 1)
        return(pval)
    LRT_pvals = []
    up_v_down_vals = []
    for column in marker_exp:
        log_reg_in = pd.DataFrame(data=[marker_exp[column]])
        log_reg_in = np.transpose(log_reg_in)
        c_list_2 = np.array(c_list)
        c_list_2 = np.array(c_list_2 == coi,dtype=int)
        c_list_2 = np.transpose(c_list_2)
        log_reg_in['cluster'] = c_list_2
        in_cls = marker_exp[column][c_list == coi].values
        out_cls = marker_exp[column][c_list != coi].values
        out_cls_mean = np.sum(out_cls) / len(out_cls)
        in_cls_mean = np.sum(in_cls) / len(in_cls)
        test = in_cls_mean - out_cls_mean
        if test <= 0:
            up_v_down_vals.append('down')
        else:
            up_v_down_vals.append('up')
        
        LRT_pval = LRT_LogReg(log_reg_in)
        LRT_pvals.append(LRT_pval)
        
    t = marker_exp.apply(
        lambda col:
        ss.ttest_ind(
            col[c_list == coi],
            col[c_list != coi],
            equal_var=False
        )
    )
    ws = marker_exp.apply(
        lambda col:
        ranksums(
            col[c_list == coi],
            col[c_list != coi]
        )
    )
    output = pd.DataFrame()
    output['gene_1'] = t.index
    #output['gene_1'] = ws.index
    output[['t_stat', 't_pval']] = pd.DataFrame(
        t.values.tolist(),
        columns=['t_stat', 't_pval']
    )
    output[['w_stat', 'w_pval']] = pd.DataFrame(
        ws.values.tolist(),
        columns=['w_stat', 'w_pval']
    )

    output['up_down'] = up_v_down_vals
    output['LRT_pval'] = LRT_pvals
    return output


def batch_stats(marker_exp, c_list, coi):
    """Applies t test & wilcoxon rank sum test to a gene expression matrix, gene by gene.

    :param marker_exp: A DataFrame whose rows are cell identifiers, columns are
        gene identifiers, and values are float values representing gene
        expression.
    :param c_list: A Series whose indices are cell identifiers, and whose
        values are the cluster which that cell is part of.
    :param coi: The cluster of interest.

    :returns: A matrix with arbitary row indices whose columns are the gene, t
              statistic, then t p-value; the last two being of float type.
              Their names are 'gene', 't_stat' , 't_pval' , 'w_stat' , 'w_pval'

    :rtype: pandas.DataFrame
    """
    ttest_result = ss.ttest_ind( marker_exp[c_list == coi], marker_exp[c_list != coi], equal_var=False)
    ws_result    = ranksums(  marker_exp[c_list == coi], marker_exp[c_list != coi])
    output = pd.DataFrame({'t_stat':ttest_result.statistic,
                           't_pval':ttest_result.pvalue,
                           'ws_stat':ws_result.statistic,
                           'ws_pval':ws_result.pvalue}, index=marker_exp.columns)
    output.index.name = 'gene_1'
        
    return output


def batch_fold_change(marker_exp, c_list, coi):
    """Applies log2 fold change to a gene expression matrix, gene by gene.

    :param marker_exp: A DataFrame whose rows are cell identifiers, columns are
        gene identifiers, and values are float values representing gene
        expression.
    :param c_list: A Series whose indices are cell identifiers, and whose
        values are the cluster which that cell is part of.
    :param coi: The cluster of interest.
    :rtype: pandas.DataFrame
    """
    
    def fold_change(col,c_list,coi):
        FC = col.groupby(c_list==coi).mean()
        mean0 = FC[True] + .000001
        mean1 = FC[False] + .000001

        if mean0 == 0:
            val = math.nan
            return val
        if mean1 == 0:
            val = math.nan
            return val
        val = (mean0/mean1)
        if val < 0:
            return abs(val)
        return val

    
    fc = marker_exp.apply(
        lambda col:
        (math.log(fold_change(col,c_list,coi),2))
        )
    fca = abs(fc)
    output = pd.DataFrame()
    output['gene_1'] = fc.index
    output[['Log2FoldChange']] = pd.DataFrame(
        fc.values.tolist(),
        columns=['Log2FoldChange']
    )
    output['gene_1'] = fca.index
    output[['Log2FoldChangeAbs']] = pd.DataFrame(
        fca.values.tolist(),
        columns=['Log2FoldChangeAbs']
    )
    return output


def mhg_cutoff_value(marker_exp, cutoff_ind):
    """Finds discrete expression cutoff value, from given cutoff index.

    The XL-mHG test outputs the index of the cutoff of highest significance
    between a sample and population.  This functions finds the expression value
    which corresponds to this index.  Cells above this value we define as
    expressing, and cells below this value we define as non-expressing.  We
    therefore choose this value to be between the expression at the index, and
    the expression of the "next-highest" cell.  I.e. for expression [3.0 3.0
    1.5 1.0 1.0] and index 4, we should choose a cutoff between 1 and 1.5. This
    implementation will add epsilon to the lower bound (i.e. the value of
    FLOAT_PRECISION).  In our example, the output will be 1.0 +
    FLOAT_PRECISION.  For FLOAT_PRECISION = 0.001, this is 1.001.

    :param marker_exp: A DataFrame whose rows are cell identifiers, columns are
        gene identifiers, and values are float values representing gene
        expression.
    :param cutoff_ind: A DataFrame whose 'gene' column are gene identifiers,
        and whose 'mHG_cutoff' column are cutoff indices

    :returns: A DataFrame whose 'gene' column are gene identifiers, and whose
              'cutoff_val' column are cutoff values corresponding to input
              cutoff indices.

    :rtype: pandas.DataFrame
    """

    def find_val(row):
        gene = row['gene_1']
        val = marker_exp[gene].sort_values(
            ascending=False).iloc[row['mHG_cutoff']]
        if re.compile(".*_negation$").match(gene):
            return val + FLOAT_PRECISION
        else:
            return val + FLOAT_PRECISION

    cutoff_ind.index = cutoff_ind['gene_1']
    cutoff_val = cutoff_ind.apply(
        find_val, axis='columns'
    ).rename('cutoff_val')
    output = cutoff_val.to_frame().reset_index()
    return output


def mhg_slide(marker_exp, cutoff_val):
    """Slides cutoff indices in XL-mHG output out of uniform expression groups.

    The XL-mHG test may place a cutoff index that "cuts" across a group of
    uniform expression inside the sorted expression list.  I.e. for a
    population of cells of which many have zero expression, the XL-mHG test may
    demand that we sample some of the zero-expression cells and not others.
    This is impossible because the cells are effectively identical.  This
    function therefore moves the XL-mHG cutoff index so that it falls on a
    measurable gene expression boundary.

    Example: for a sorted gene expression list [5, 4, 1, 0, 0, 0] and XL-mHG
    cutoff index 4, this function will "slide" the index to 3; marking the
    boundary between zero expression and expression value 1.

    :param marker_exp: A DataFrame whose rows are cell identifiers, columns are
        gene identifiers, and values are float values representing gene
        expression.
    :param cutoff_val: A DataFrame whose 'gene' column are gene identifiers,
        and whose 'cutoff_val' column are cutoff values corresponding to input
        cutoff indices.

    :returns: A DataFrame with 'gene', 'mHG_cutoff', and 'cutoff_val' columns,
              slid.

    :rtype: pandas.DataFrame
    """
    '''
    def searcher(row):
        if re.compile(".*_negation$").match(row['gene_1']):
            val_ =  np.searchsorted(
                -marker_exp[row['gene_1']].sort_values(ascending=False).values,
                -row['cutoff_val'], side='left')
        else:
            val_ = np.searchsorted(
                -marker_exp[row['gene_1']].sort_values(ascending=False).values,
                -row['cutoff_val'], side='left')
        
        return val_
        
    cutoff_val.index = cutoff_val['gene_1']
    cutoff_ind = cutoff_val.apply(lambda row: searcher(row),axis='columns')
    '''
    cutoff_val.index = cutoff_val['gene_1']
    cutoff_ind = cutoff_val.apply(
        lambda row:
        np.searchsorted(
            -marker_exp[row['gene_1']].sort_values(ascending=False).values,
            -row['cutoff_val'], side='left'
        ),
        axis='columns'
    )
    output = cutoff_val
    output['mHG_cutoff'] = cutoff_ind
    # Reorder and remove redundant row index
    output = output.reset_index(
        drop=True)[['gene_1', 'mHG_cutoff', 'cutoff_val']]
    return output


def discrete_exp(marker_exp, cutoff_val,abbrev,xlmhg):
    """Converts a continuous gene expression matrix to discrete.

    As a note: cutoff values correspond to the "top" of non-expression.  Only
    cells expressing at values greater than the cutoff are marked as
    "expressing"; cells expressing at the cutoff exactly are not.

    :param marker_exp: A DataFrame whose rows are cell identifiers, columns are
        gene identifiers, and values are float values representing gene
        expression.
    :param cutoff_val: A Series whose rows are gene identifiers, and values are
        cutoff values.

    :returns: A gene expression matrix identical to marker_exp, but with
              boolean rather than float expression values.

    :rtype: pandas.DataFrame
    """
    output = (marker_exp>cutoff_val).astype('int32')

    ab = '2'
    abb = '3'
    if ab in abbrev or abb in abbrev:
        #Ensures the matrix is in the proper order to get the top however many genes for the heuristic approach
        #create 'new_order', a list with the new ordering. trips_list goes first, rest after in any order
        rest = []
        new_order = xlmhg['gene_1'].tolist()
        output = output[new_order]
    return output


def tp_tn(discrete_exp, c_list, coi, cluster_overall):
    """Finds simple true positive/true negative values for the cluster of
    interest.

    :param discrete_exp: A DataFrame whose rows are cell identifiers, columns
        are gene identifiers, and values are boolean values representing gene
        expression.
    :param c_list: A Series whose indices are cell identifiers, and whose
        values are the cluster which that cell is part of.
    :param coi: The cluster of interest.

    :returns: A matrix with arbitary row indices, and has 3 columns: one for
              gene name, then 2 containing the true positive and true negative
              values respectively.  Their names are 'gene', 'TP', and 'TN'.

    :rtype: pandas.DataFrame
    """

    #does rest of clusters
    discrete_exp.fillna(0, inplace=True)
    sing_cluster_exp_matrices = {}
    for clstrs in cluster_overall:
        if clstrs == coi:
            continue
        sing_cluster_exp_matrices[clstrs] = discrete_exp.groupby(c_list == clstrs).mean()
        sing_cluster_exp_matrices[clstrs] = sing_cluster_exp_matrices[clstrs].T.rename(columns={True:'TP',False:'TN'})
        sing_cluster_exp_matrices[clstrs]['TN'] = 1 - sing_cluster_exp_matrices[clstrs]['TN']
        sing_cluster_exp_matrices[clstrs].index.name = 'gene_1'

    output = discrete_exp.groupby(c_list == coi).mean().T.rename(columns={True:'TP',False:'TN'})
    output['TN'] = 1 - output['TN']
    output.index.name = 'gene_1'
    output.reset_index(inplace=True)
    output.rename_axis('gene_1')

    #outputs a DF for COI and a dict of DF's for rest
    return output, sing_cluster_exp_matrices


def pair_product(discrete_exp, c_list, coi, cluster_number,cluster_overall):
    """Finds paired expression counts.  Returns in matrix form.

    (Number of cells in which two genes are coexpressed)

    The product of the transpose of the discrete_exp DataFrame is a matrix
    whose rows and columns correspond to individual genes.  Each value is the
    number of cells which express both genes (i.e. the dot product of two lists
    of 1s and 0s encoding expression/nonexpression for their respective genes
    in the population).  The product therefore encodes joint expression counts
    for any possible gene pair (including a single gene paired with itself).

    This function produces two matrices: one considering only cells inside the
    cluster of interest, and one considering all cells in the population.

    This function also produces a list mapping integer indices to gene names,
    and the population cell count.

    Additionally, only the upper triangular part of the output matrices is
    unique.  This function therefore also returns the upper triangular indices
    for use by other functions; this is a lazy workaround for the issue that
    comes with using columns 'gene_1' and 'gene_2' to store gene pairs; the
    gene pair (A, B) is therefore treated differently than (B, A).  Specifying
    the upper triangular part prevents (B, A) from existing.

    TODO: fix this redundancy using multi-indices

    :param discrete_exp: A DataFrame whose rows are cell identifiers, columns
        are gene identifiers, and values are boolean values representing gene
        expression.
    :param c_list: A Series whose indices are cell identifiers, and whose
        values are the cluster which that cell is part of.
    :param coi: The cluster of interest.
    :param clusters_number: dict with clusters and their sizes

    :returns: (gene mapping list, cluster count, total count, cluster paired
              expression count matrix, population paired expression count
              matrix, upper triangular matrix index)

    :rtype: (pandas.Index, int, int, numpy.ndarray, numpy.ndarray,
            numpy.ndarray)
    """

    '''
    ###############
    #Heuristic Section
    if trips_list == None:
        pass
    else:
        for column in discrete_exp:
            #print(column)
            if str(column) in trips_list:
                continue
            else:
                discrete_exp.drop(column, axis=1,inplace=True)
    ################
    '''
    
    gene_map = discrete_exp.columns
    in_cls_matrix = discrete_exp[c_list == coi].values
    total_matrix = discrete_exp.values
    #get exp matrices for each non-interest cluster for new rank scheme
    cluster_exp_matrices = {}
    cls_counts = {}
    for clstrs in cluster_overall:
        if clstrs == coi:
            pass
        else:
            cluster_exp_matrices[clstrs] = discrete_exp[c_list == (clstrs)].values
            cls_counts[clstrs] = np.size(cluster_exp_matrices[clstrs],0)
            cluster_exp_matrices[clstrs]=np.matmul(np.transpose(cluster_exp_matrices[clstrs]),cluster_exp_matrices[clstrs] )

    in_cls_count = np.size(in_cls_matrix, 0)
    pop_count = np.size(total_matrix, 0)
    in_cls_product = np.matmul(np.transpose(in_cls_matrix), in_cls_matrix)
    total_product = np.matmul(np.transpose(total_matrix), total_matrix)
    upper_tri_indices = np.triu_indices(gene_map.size,1)
    
    return (
        gene_map,
        in_cls_count, pop_count,
        in_cls_product, total_product,
        upper_tri_indices, cluster_exp_matrices, cls_counts
    )

  
    

def combination_product(discrete_exp,c_list,coi,abbrev,heur_limit):
    '''
    Makes the count matrix for combinations of K genes where K > 2 :)

    X -> in_cls_matrix

    First constructs the extended gene by cell matrix then multiplies by the transpose
    -> (X^T & X^T) * X
    & represents AND bitwise operation to construct the gene combos by cells matrix
    (need to perform AND on each row in first X^T by each row in second X^T)
    This scales to K genes , just need to do more ANDs with more X^Ts

    Once the first step is complete, simply multiply using np.matmul to get the gene by gene
    matrix. Need to also make a new gene_map

    More in-depth Description:
    the trips matrix is a gene/gene pair by gene matrix. TO construct, we first perform an AND 
    operation on each given row with a different row from the same matrix(X&X), we use two 
    identical matrices and loop through most combinations. The code SKIPS any entry that would be
    the same pair (e.g. AA) and also doesn't do inverses of already counted pairs
    (e.g. do AB, dont do BA). this gives us a somewhat unique looking rectangular matrix that isnt
    quit N^2 long. From here, we  multiply by the original matrix to get our full gene expression
    count matrix. To be absolutely clear, the matrices we use to construct the ~N^2 gene pair 
    matrix are actually the original discrete expression matrix transposed so that the genes
    and cells end up in the right place. We do this for our cluster and the total population,
    just as in the pair case. After this, we need to make a smarter gene mapping since the
    terms are not as easy to predict as in a square matrix. To do this, a pattern was determined
    (specifically, the pairs follow the N'th triangular number scheme), so from there we construct
    the full mapping for genes 1 & 2 to be used later in the data table. 'Trips_indices' gives us
    the indices for the entire rectangular matrix, something to feed into the vectorized HG & TPTN
    when we get there. Gene 3 mapping follows these indices so there is no separate mapping.

    '''

    def trips_matrix_gen_heuristic(matrix,heur_limit):
        #transpose gives us genes as rows
        Cth = time.time()
        trips_matrix_list = []
        row1count = 1
        for row1 in np.transpose(matrix)[:heur_limit]:
            row2count = 1
            for row2 in np.transpose(matrix):
                if row2count<=row1count:
                    row2count = row2count+1
                    continue
                trips_matrix_list.append(row1&row2)
                row2count=row2count+1
            row1count = row1count + 1
        #trips_matrix is in cluster combo gene by cell
        Ath = time.time()
        trips_matrix = np.asarray(trips_matrix_list)
        trips_product = np.matmul(trips_matrix,np.transpose(np.transpose(matrix)[:heur_limit]))
        Bth = time.time()
        print(str(Bth-Ath) + ' mult seconds')
        return trips_product
    

    def trips_matrix_gen(matrix):
        Cth = time.time()
        trips_matrix_list = []
        row1count = 1
        for row1 in np.transpose(matrix):
            row2count = 1
            for row2 in np.transpose(matrix):
                if row2count<=row1count:
                    row2count = row2count+1
                    continue
                trips_matrix_list.append(row1&row2)
                row2count=row2count+1
            row1count = row1count + 1

        #trips_matrix is in cluster combo gene by cell
        Ath = time.time()
        trips_matrix = np.asarray(trips_matrix_list)
        trips_product = np.matmul(trips_matrix,matrix)
        Bth = time.time()
        print(str(Bth-Ath) + ' mult seconds')
        return trips_product

    
    
    start_time = time.time()
    in_cls_matrix = discrete_exp[c_list == coi].values
    total_matrix = discrete_exp.values
    gene_count = len(discrete_exp.columns)
    first = time.time()
    if '3' in abbrev:
        total_pairs =  0
        for i in range(1,heur_limit+1):
            total_pairs = total_pairs + (len(discrete_exp.columns.values) - i)
        total_pairs = total_pairs * len(discrete_exp.columns.values)
        trips_in_cls_product = trips_matrix_gen_heuristic(in_cls_matrix,heur_limit)
        trips_total_product = trips_matrix_gen_heuristic(total_matrix,heur_limit)
    else:
        trips_in_cls_product = trips_matrix_gen(in_cls_matrix)
        trips_total_product = trips_matrix_gen(total_matrix)

    #make a row-wise gene_map scheme
    gene_map = discrete_exp.columns.values
    gene_1_mapped = []
    count = 0
    count_2 = 0
    for gene in gene_map:
        if '3' in abbrev and count_2 >= total_pairs+1:
            break
        for x in range(gene_count-1-count):
            for n in range(gene_count):
                gene_1_mapped.append(gene)
                count_2 = count_2 + 1
        count = count + 1
    gene_1_mapped = pd.Index(gene_1_mapped)
    if '3' in abbrev:
        gene_1_mapped = gene_1_mapped[:total_pairs]
    count_2 = 0
    gene_3_mapped = []
    val = int(len(gene_1_mapped)/gene_count)
    for x in range(val):
        if '3' in abbrev and count_2 >= total_pairs+1:
            break
        count = 0
        for gene in gene_map:
            gene_3_mapped.append(gene_map[count])
            count_2 = count_2 + 1
            count = count + 1
    gene_3_mapped = pd.Index(gene_3_mapped)
    if '3' in abbrev:
        gene_3_mapped = gene_3_mapped[:total_pairs]
    count_2 = 0
    gene_2_mapped = []
    for x in range(gene_count-1):
        if '3' in abbrev and count_2 >= total_pairs+1:
            break
        count = 0
        for gene in gene_map:
            if count == 0:
                count = count + 1
                continue
            for n in range(gene_count):
                gene_2_mapped.append(gene_map[count])
                count_2 = count_2 + 1
            count = count + 1
        gene_map = np.delete(gene_map,0)
    gene_2_mapped = pd.Index(gene_2_mapped)
    if '3' in abbrev:
        gene_2_mapped = gene_2_mapped[:total_pairs]

    #indices for computation
    row_count = int((gene_count*(gene_count-1))/2)
    
    #column coordinate
    count_2 = 0
    list_one_ = []
    for numm in range(row_count):
        if '3' in abbrev and count_2 >= total_pairs+1:
            break
        for num in range(gene_count):
            list_one_.append(num)
            count_2 = count_2 + 1
    list_one = np.array(list_one_)
    
    #row coordinate
    count_2 = 0
    list_two_ = []
    for num in range(row_count):
        if '3' in abbrev and count_2 >= total_pairs+1:
            break
        for numm in range(gene_count):
            list_two_.append(num)
            count_2 = count_2 + 1
    list_two = np.array(list_two_)
    
    fourth = time.time()

    if '3' in abbrev:
        list_one = list_one[:total_pairs]
        list_two = list_two[:total_pairs]
        
    trips_indices = ( list_two , list_one )
    
    return (
        trips_in_cls_product,trips_total_product,trips_indices,
        gene_1_mapped,gene_2_mapped,gene_3_mapped
        )


def pair_hg(gene_map, in_cls_count, pop_count, in_cls_product, total_product,
            upper_tri_indices, abbrev,heur_limit):
    """Finds hypergeometric p-value of gene pairs.

    Takes in discrete single-gene expression matrix, and finds the
    hypergeometric p-value of the sample that includes cells which express both
    of a pair of genes. 

    hypergeometric parameters:
    k = in_cls_product (number of cells in cluster with coexpression of both genes in the given pair)
    M = pop_count (total number of cells)
    n = in_cls_count (number of cells in cluster)
    N = total_product  (total number of cells with coexpression of both genes in the given pair)

    :param gene_map: An Index mapping index values to gene names.
    :param in_cls_count: The number of cells in the cluster.
    :param pop_count: The number of cells in the population.
    :param in_cls_product: The cluster paired expression count matrix.
    :param total_product: The population paired expression count matrix.
    :param upper_tri_indices: An array specifying UT indices; from numpy.utri

    :returns: A matrix with columns: the two genes of the pair, hypergeometric
              test statistics for that pair.  Their names are 'gene_1',
              'gene_2', 'HG_stat'.

    :rtype: pandas.DataFrame

    """
    
    vhg = np.vectorize(ss.hypergeom.sf, excluded=[1, 2, 4], otypes=[np.float])
    ab = '2'
    if ab in abbrev:
        count = 0
        row_count = 1
        heuristic_limit = heur_limit
        for num in range(heuristic_limit):
            count = count + (len(gene_map) - row_count)
            row_count = row_count + 1
        revised_row_indices = np.asarray(upper_tri_indices[0][:count])
        revised_col_indices = np.asarray(upper_tri_indices[1][:count])
        revised_indices = (revised_row_indices,revised_col_indices)
        # Only apply to upper triangular    
        hg_result = vhg(
            in_cls_product[revised_indices],
            pop_count,
            in_cls_count,
            total_product[revised_indices],
            loc=1
            )
        output = pd.DataFrame({
            'gene_1': gene_map[revised_indices[0]],
            'gene_2': gene_map[revised_indices[1]],
            'HG_pval': hg_result
            }, columns=['gene_1', 'gene_2', 'HG_pval'])
        return output, revised_indices

    
    else:
        hg_result = vhg(
            in_cls_product[upper_tri_indices],
            pop_count,
            in_cls_count,
            total_product[upper_tri_indices],
            loc=1
        )
        output = pd.DataFrame({
            'gene_1': gene_map[upper_tri_indices[0]],
            'gene_2': gene_map[upper_tri_indices[1]],
            'HG_pval': hg_result
        }, columns=['gene_1', 'gene_2', 'HG_pval'])
        return output, None



def ranker(pair,xlmhg,sing_tp_tn,other_sing_tp_tn,other_pair_tp_tn,cls_counts,in_cls_count,pop_count):
    """
    :param pair: table w/ gene_1, gene_2, HG_stat as columns (DataFrame (DF) )
    :param xlmhg: DF of mHG stats for each gene(for testing lead vs follow gene in pair)
    :param other_sing_tp_tn: TP/TN values for singletons in all other clusters (dict of DFs)
    :param other_pair_tp_tn: TP/TN values for pairs in all other clusters (dict of DFs)
    :param cls_counts: # of cells in a given cluster (dict of DFs)
    **
    All dicts have style: 
    key -> cluster number
    value -> data
    **

    Statistic to calculate is :
    SUM across all clusters of (TN_after - TN_before) / N
    where:
    TN_after = TN of gene combo in cluster
    TN_before = TN of initial gene from pair in cluster
    N = # of cells in the cluster
    
    THRESHOLDING:
    -We are only taking the first 100 appearances for each gene in all the pairs
    -'Lead' gene is one with smallest p val

    returns: New pair table w/ new columns and ranks.
    ranked-pair is a DEEP copy of pair, meaning value changes in it
    are not reflected in pair 
    """

    def ranked_stat(gene_1,gene_2,lead_gene,cls_counts,other_pair_tp_tn,other_sing_tp_tn,in_cls_count):
        stats=[]
        MGDstats=[]
        stats_debug = {}
        for clstrs in cls_counts:
            TN_before = other_sing_tp_tn[clstrs].at[lead_gene,'TN']
            TN_after = other_pair_tp_tn[clstrs].at[(gene_1,gene_2),'TN']
            N = cls_counts[clstrs]
            #value =  ( TN_after - TN_before ) / N
            value =  ( TN_after - TN_before )
            stats.append(value)
            MGDstats.append(((TN_after - TN_before)*N)/pop_count)
        stat1 = sum(stats)
        MGDstat = sum(MGDstats)
        #stats.remove(stat1)
        #stat2 = max(stats)
        #stat= stat1+stat2
        return stat1,MGDstat

    
    xlmhg_cop = xlmhg.copy()
    xlmhg_cop = xlmhg_cop.set_index('gene_1')
    ranked_pair = pair.copy()
    num_tests = len(ranked_pair)
    ranked_pair.sort_values(by='HG_pval',ascending=True,inplace=True)
    ranked_pair['HG_rank'] = ranked_pair.reset_index().index + 1
    
    #below not used because this does ALL pairs (too many)
    #ranked_pair['CCS'] = ranked_pair.apply(ranked_stat,axis=1,args=(cls_counts,other_pair_tp_tn,other_sing_tp_tn))
    omit_pairs = {}
    count = 1
    if len(ranked_pair.index) < 5000:
        thresh = 1000
    else:
        thresh = 1000
    loopstart = time.time()
    
    for index,row in ranked_pair.iterrows():
        if row[0] in omit_pairs:
            if omit_pairs[row[0]] > 200:
                ranked_pair.drop(index, inplace=True)
                continue
        if row[1] in omit_pairs:
            if omit_pairs[row[1]] > 200:
                ranked_pair.drop(index, inplace=True)
                continue
        
        gene_1 = row[0]
        gene_2 = row[1]
        #determine which of the two genes is 'lead'
        stat_1 = xlmhg_cop.at[gene_1,'mHG_pval']
        stat_2 = xlmhg_cop.at[gene_2,'mHG_pval']
        if stat_1 <= stat_2:
            lead_gene=gene_1
            follow_gene = gene_2
        else:
            lead_gene=gene_2
            follow_gene = gene_1


        if row[0] in omit_pairs:
            omit_pairs[row[0]] = omit_pairs[row[0]] + 1
        if row[1] in omit_pairs:
            omit_pairs[row[1]] = omit_pairs[row[1]] + 1
            
        if row[0] not in omit_pairs:
            omit_pairs[row[0]] = 1
        if row[1] not in omit_pairs:
            omit_pairs[row[1]] = 1
        
        if count == thresh:
            break
        
        ranked_pair.at[index,'CCS'],ranked_pair.at[index,'MGD'] = ranked_stat(gene_1,gene_2,lead_gene,cls_counts,other_pair_tp_tn,other_sing_tp_tn,in_cls_count)
        count = count + 1

    loopend = time.time()
    ranked_pair.sort_values(by='CCS',ascending=False,inplace=True)
    ranked_pair['CCS_rank'] = ranked_pair.reset_index().index + 1
    ranked_pair['finalrank'] = ranked_pair[['HG_rank', 'CCS_rank']].mean(axis=1)
    ranked_pair.sort_values(by='finalrank',ascending=True,inplace=True)
    ranked_pair['rank'] = ranked_pair.reset_index().index + 1
    ranked_pair.drop('finalrank',axis=1,inplace=True)
    omit_genes_lead = {}
    omit_genes_follow = {}
    count = 0
    if len(ranked_pair.index) < 5000:
        plot_num = 50
    else:
        plot_num = 100
    for index,row in ranked_pair.iterrows():
        lead_on = 1
        follow_on = 1
        if count == plot_num:
            break
        #If a lead gene has appeared more than 10 times, dont plot anymore
        if row[0] in omit_genes_lead:
            if omit_genes_lead[row[0]] >= 10:
                lead_on = 0
                #ranked_pair.at[index,'Plot'] = 0
            else:
                # test if the non-adjusted pvalue is less than .05
                if row[2]  >= .05:
                    lead_on = 0
                    #ranked_pair.at[index,'Plot'] = 0
                else:
                    #If true positive rate is less than 15%, dont plot
                    if row[3] <= .15:
                        lead_on = 0
                        #ranked_pair.at[index,'Plot'] = 0
                    else:
                        pass
                        #ranked_pair.at[index,'Plot'] = 1
                omit_genes_lead[row[0]] = omit_genes_lead[row[0]]+1
        else:
            omit_genes_lead[row[0]] = 1
            if row[2] >= .05:
                lead_on = 0
                #ranked_pair.at[index,'Plot'] = 0
            else:
                pass
                #ranked_pair.at[index,'Plot'] = 1
        ##LEAD GENE END##
        
        # Follow Gene Start #
        if row[1] in omit_genes_follow:
            if omit_genes_follow[row[1]] >= 20:
                follow_on = 0
                #ranked_pair.at[index,'Plot'] = 0
            else:
                # test if the non-adjusted pvalue is less than .05
                if row[2]  >= .05:
                    follow_on = 0
                    #ranked_pair.at[index,'Plot'] = 0
                else:
                    #If true positive rate is less than 15%, dont plot
                    if row[3] <= .15:
                        follow_on = 0
                        #ranked_pair.at[index,'Plot'] = 0
                    else:
                        pass
                        #ranked_pair.at[index,'Plot'] = 1
                omit_genes_follow[row[1]] = omit_genes_follow[row[1]]+1
        else:
            omit_genes_follow[row[1]] = 1
            if row[2] >= .05:
                lead_on = 0
                #ranked_pair.at[index,'Plot'] = 0
            else:
                pass
                #ranked_pair.at[index,'Plot'] = 1
        
        ##FOLLOW GENE END##
        if follow_on == 1 and lead_on == 1:
            ranked_pair.at[index,'Plot'] = 1
            count = count + 1
        else:
            ranked_pair.at[index,'Plot'] = 0
        
    pop_list = []
    for key, value in omit_pairs.items():
        if value <= 20:
            pop_list.append(key)

    for item in pop_list:
        omit_pairs.pop(item)
    return ranked_pair,omit_pairs


def pair_tp_tn(gene_map, in_cls_count, pop_count, in_cls_product,
               total_product, upper_tri_indices, abbrev, revised_indices):
    """Finds simple true positive/true negative values for the cluster of
    interest, for all possible pairs of genes.

    :param gene_map: An Index mapping index values to gene names.
    :param in_cls_count: The number of cells in the cluster.
    :param pop_count: The number of cells in the population.
    :param in_cls_product: The cluster paired expression count matrix.
    :param total_product: The population paired expression count matrix.
    :param upper_tri_indices: An array specifying UT indices; from numpy.utri
    :param cluster_exp_matrices: dict containing expression matrices of 
         all clusters except the cluster of interest

    :returns: A matrix with arbitary row indices and 4 columns: containing the
              two genes of the pair, then true positive and true negative
              values respectively.  Their names are 'gene_1', 'gene_2', 'TP',
              and 'TN'.

    :rtype: pandas.DataFrame

    """


    def tp(taken_in_cls):
        return taken_in_cls / in_cls_count

    def tn(taken_in_cls, taken_in_pop):
        return (
            ((pop_count - in_cls_count) - (taken_in_pop - taken_in_cls))
            / (pop_count - in_cls_count)
        )

    ab = '2'
    if ab in abbrev:
        tp_result = np.vectorize(tp)(in_cls_product[revised_indices])
        tn_result = np.vectorize(tn)(
            in_cls_product[revised_indices], total_product[revised_indices]
            )

    
        output = pd.DataFrame({
            'gene_1': gene_map[revised_indices[0]],
            'gene_2': gene_map[revised_indices[1]],
            'TP': tp_result,
            'TN': tn_result
            }, columns=['gene_1', 'gene_2', 'TP', 'TN'])

        return output
        
    else:
        tp_result = np.vectorize(tp)(in_cls_product[upper_tri_indices])
        tn_result = np.vectorize(tn)(
            in_cls_product[upper_tri_indices], total_product[upper_tri_indices]
            )
        output = pd.DataFrame({
            'gene_1': gene_map[upper_tri_indices[0]],
            'gene_2': gene_map[upper_tri_indices[1]],
            'TP': tp_result,
            'TN': tn_result
            }, columns=['gene_1', 'gene_2', 'TP', 'TN'])

        return output



def trips_hg(gene_map,in_cls_count,pop_count,trips_in_cls,trips_total,trips_indices,gene_1_mapped,gene_2_mapped,gene_3_mapped,abbrev,heur_limit):
    '''
    Altered version of pair_hg to do trips w/ the trips expression matrices
    See pair_hg for general var descriptions
    See combination product for full trips discussion
    Uses same vectorization scheme as pair, but has new rectangular indices and 
    third gene output. 'pair_indices' just gives us a list from 0 to however many values
    there are, this is then checked against the gene maps which contain all the mapping
    info for genes 1 & 2 (this setup plays nicely with the dataframe construction, the gene map
    values are already in the correct order). We then trim the matrix, disallowing any values 
    that have repeat genes (e.g. ACC -> no good) and also removing repeat triplets
    (e.g. ABC -> good , BAC -> no good if ABC already exists). This is SLOW, but more or less
    unavoidable. We didn't need to do this in the pair case b/c we knew ahead of time that all
    valid and unique values would be in the upper triangle, that doesn't hold in this case.
    '''


    def tp(taken_in_cls):
        return taken_in_cls / in_cls_count

    def tn(taken_in_cls, taken_in_pop):
        return (
            ((pop_count - in_cls_count) - (taken_in_pop - taken_in_cls))
            / (pop_count - in_cls_count)
        )


    abb = '3'
    if abb in abbrev:
        new_row_indices = []
        new_col_indices = []
        new_gene_1_mapped = []
        new_gene_2_mapped = []
        new_gene_3_mapped = []
        num_genes = len(gene_map)
        heuristic_limit = heur_limit


        #reassign all gene_maps and indices of computation for hg test
        #count -> position in the row to cut off the right half of the matrix
        #Need to cut out the mappings from the right half of the matrix

            
        #new col indices
        count = 0
        for num in trips_indices[1]:
            if count <= heuristic_limit-1:
                new_col_indices.append(num)
            elif count > heuristic_limit-1 and count < len(gene_map)-1:
                pass
            else:
                count = 0
                continue
            count = count + 1
        #new row indices
        count = 0
        for num in trips_indices[0]:
            if count <= heuristic_limit-1:
                new_row_indices.append(num)
            elif count > heuristic_limit-1 and count < len(gene_map)-1:
                pass
            else:
                count = 0
                continue
            count = count + 1
        #New gene 1 mapping
        count = 0
        for gene in gene_1_mapped:
            if count <= heuristic_limit-1:
                new_gene_1_mapped.append(gene)
            elif count > heuristic_limit-1 and count < len(gene_map)-1:
                pass
            else:
                count = 0
                continue
            count = count + 1
        #New gene 2 mapping 
        count = 0
        for gene in gene_2_mapped:
            if count <= heuristic_limit-1:
                new_gene_2_mapped.append(gene)
            elif count > heuristic_limit-1 and count < len(gene_map)-1:
                pass
            else:
                count = 0
                continue
            count = count + 1
        #New gene 3 mapping
        count = 0
        for gene in gene_3_mapped:
            if count <= heuristic_limit-1:
                new_gene_3_mapped.append(gene)
            elif count > heuristic_limit-1 and count < len(gene_map)-1:
                pass
            else:
                count = 0
                continue
            count = count + 1

        new_gene_1_mapped = pd.Index(new_gene_1_mapped)
        new_gene_2_mapped = pd.Index(new_gene_2_mapped)
        new_gene_3_mapped = pd.Index(new_gene_3_mapped)
        revised_row_indices = np.asarray(new_row_indices)
        revised_col_indices = np.asarray(new_col_indices)
        revised_indices = (revised_row_indices,revised_col_indices)

        
        tp_result = np.vectorize(tp)(trips_in_cls[revised_indices])
        tn_result = np.vectorize(tn)(
            trips_in_cls[revised_indices], trips_total[revised_indices]
        )
    
        vhg = np.vectorize(ss.hypergeom.sf, excluded=[1, 2, 4], otypes=[np.float])

    
        hg_result = vhg(
            trips_in_cls[revised_indices],
            pop_count,
            in_cls_count,
            trips_total[revised_indices],
            loc=1
        )
        print('HG + TP/TN done')
        pair_indices = []
        gene_count = int(len(gene_map))
        val_count =int( len(new_gene_1_mapped))
        for x in range(val_count):
            pair_indices.append(x)
        pair_indices = np.array(pair_indices)
        
        output = pd.DataFrame({
            'gene_1': new_gene_1_mapped[pair_indices],
            'gene_2': new_gene_2_mapped[pair_indices],
            'gene_3': new_gene_3_mapped[pair_indices],
            'HG_stat': hg_result,
            'TP' : tp_result,
            'TN': tn_result
        }, columns=['gene_1', 'gene_2', 'gene_3', 'HG_stat','TP','TN'])

    else:
        tp_result = np.vectorize(tp)(trips_in_cls[trips_indices])
        tn_result = np.vectorize(tn)(
            trips_in_cls[trips_indices], trips_total[trips_indices]
        )
    
        vhg = np.vectorize(ss.hypergeom.sf, excluded=[1, 2, 4], otypes=[np.float])

    
        hg_result = vhg(
            trips_in_cls[trips_indices],
            pop_count,
            in_cls_count,
            trips_total[trips_indices],
            loc=1
        )

        print('HG + TP/TN done')
        pair_indices = []
        gene_count = int(len(gene_map))
        val_count =int( len(gene_1_mapped))
        for x in range(val_count):
            pair_indices.append(x)
        pair_indices = np.array(pair_indices)
        output = pd.DataFrame({
            'gene_1': gene_1_mapped[pair_indices],
            'gene_2': gene_2_mapped[pair_indices],
            'gene_3': gene_3_mapped[pair_indices],
            'HG_stat': hg_result,
            'TP' : tp_result,
            'TN': tn_result
        }, columns=['gene_1', 'gene_2', 'gene_3', 'HG_stat','TP','TN'])



    
    output = output.sort_values(by='HG_stat', ascending=True)
    output = output.head(10000)
    #trims off values w/ repeating genes (e.g. ACC)
    #store unique trios in used_genes
    #iteratively check against the list
    used_genes = []
    counter=0
    prev_combo = []
    dropped=0
    buff = 0
    for index, row in output.iterrows():
        dropped = 0
        if counter == 2000:
            break
        if row[0]==row[1] or row[1]==row[2] or row[0]==row[2]:
            output.drop([index],inplace=True)
            continue
        for combo in list(reversed(prev_combo)):
            if row[0] in combo and row[1] in combo and row[2] in combo:
                output.drop([index],inplace=True)
                dropped = 1
                break
        if dropped != 1:
            prev_combo.append( [ row[0],row[1],row[2] ] )
            counter = counter + 1
    return output




from setuptools import setup
import os

def read(filename):
    return open(os.path.join(os.path.dirname(__file__), filename)).read()

setup(
    name = 'COMETSC',
    version = '0.1.13',
    long_description=read('README.md'),
    packages = ['Comet'],
    install_requires=[
        'adjustText==0.7.3' ,
        'atomicwrites==1.1.5' ,
        'attrs==18.1.0',
        'chardet==3.0.4',
        'cycler==0.10.0',
        'Cython==0.28.4',
        'decorator==4.3.0',
        'future==0.16.0',
        'idna==2.7',
        'ipython-genutils==0.2.0',
        'jsonschema==2.6.0',
        'jupyter-core==4.4.0',
        'kiwisolver==1.0.1',
        'matplotlib==3.0.0',
        'more-itertools==4.2.0',
        'nbformat==4.4.0',
        'numpy==1.14.5',
        'pandas==0.23.3',
        'plotly==2.7.0',
        'pluggy==0.6.0',
        'py==1.10.0',
        'pyparsing==2.2.0',
        'pytest==3.6.3',
        'python-dateutil==2.7.3',
        'pytz==2018.5',
        'requests==2.20.1',
        'scipy==1.1.0',
        'scikit-learn==0.21.0',
        'six==1.11.0',
        'traitlets==4.3.2',
        'urllib3==1.24.3',
        'xlmhg==2.4.9'],
    include_package_data=True,
    entry_points = {
        'console_scripts': [
            'Comet = Comet.__main__:main'
            ]
        }
    
)



import re

import pandas as pd
import numpy as np
import xlmhg as hg
import scipy.stats as ss
import time
import math

FLOAT_PRECISION = 0.001
  
    



def combination_product(discrete_exp,c_list,coi,xlmhg):
    count = 0
    trips_list=[]
    for index,row in xlmhg.iterrows():
        if count == 50:
            break
        else:
            trips_list.append(row['gene_1'])
            count = count + 1


    def quads_matrix_gen(matrix):
        Cth = time.time()
        quads_matrix_list = []
        for row1 in np.transpose(matrix):
            for row2 in np.transpose(matrix):
                quads_matrix_list.append(row1&row2)
        quads_matrix = np.asarray(quads_matrix_list)
        print('')
        print('quads matrix times itself')
        print(quads_matrix.shape)
        print('')
        Ath = time.time()
        quads_product = np.matmul(quads_matrix,np.transpose(quads_matrix))
        #print('N^2 x N^2 full count matrix hopefully')
        #print(quads_product)
        #print(quads_product.shape)
        #print('')
        Bth = time.time()
        print(str(Bth-Ath) + ' mult seconds')
        return quads_product

    ###############
    #Experimental Section
    if trips_list == None:
        pass
    else:
        for column in discrete_exp:
            #print(column)
            if str(column) in trips_list:
                continue
            else:
                discrete_exp.drop(column, axis=1,inplace=True)
    ################
    
    
    start_time = time.time()
    #print('discrete exp matrix')
    #print(discrete_exp)
    in_cls_matrix = discrete_exp[c_list == coi].values
    #print('')
    #print('in cls matrix')
    #print(in_cls_matrix)
    #print('')
    total_matrix = discrete_exp.values
    #print('total matrix')
    #print(total_matrix)
    #print('')
    gene_count = len(discrete_exp.columns)
    first = time.time()
    quads_in_cls_product = quads_matrix_gen(in_cls_matrix)
    quads_total_product = quads_matrix_gen(total_matrix)

    gene_map = discrete_exp.columns.values
    #print(gene_map)
    odd_gene_map = []
    count = 0
    for gene in gene_map:
        for x in range(gene_count):
            odd_gene_map.append(gene)
    #print(odd_gene_map)
    #print('')
    odd_gene_map = pd.Index(odd_gene_map)
    even_gene_map = []
    count = 0
    for gene in gene_map:
        for x in gene_map:
            even_gene_map.append(x)
    #print(even_gene_map)
    even_gene_map = pd.Index(even_gene_map)

        
    quads_indices = np.triu_indices((gene_map.size*gene_map.size),1)
    return (
        quads_in_cls_product,quads_total_product,quads_indices,odd_gene_map,even_gene_map
        )





def quads_hg(gene_map,in_cls_count,pop_count,quads_in_cls,quads_total,quads_indices,odd_gene_map,even_gene_map):

    def tp(taken_in_cls):
        return taken_in_cls / in_cls_count

    def tn(taken_in_cls, taken_in_pop):
        return (
            ((pop_count - in_cls_count) - (taken_in_pop - taken_in_cls))
            / (pop_count - in_cls_count)
        )
    st = time.time()
    tp_result = np.vectorize(tp)(quads_in_cls[quads_indices])
    tn_result = np.vectorize(tn)(
        quads_in_cls[quads_indices], quads_total[quads_indices]
    )
    
    vhg = np.vectorize(ss.hypergeom.sf, excluded=[1, 2, 4], otypes=[np.float])

    
    hg_result = vhg(
        quads_in_cls[quads_indices],
        pop_count,
        in_cls_count,
        quads_total[quads_indices],
        loc=1
    )
    #0th index in quads_indices refers to 0th pair of genes
    
    #print(quads_indices[0])
    #print(quads_indices[1])
    #print(odd_gene_map)
    #print(even_gene_map)
    #print(gene_map)
    #print('HG + TP/TN done')
    output = pd.DataFrame({
        'gene_1': odd_gene_map[quads_indices[0]],
        'gene_2': even_gene_map[quads_indices[0]],
        'gene_3': odd_gene_map[quads_indices[1]],
        'gene_4': even_gene_map[quads_indices[1]],
        'HG_stat': hg_result,
        'TP' : tp_result,
        'TN': tn_result
    }, columns=['gene_1', 'gene_2', 'gene_3', 'gene_4', 'HG_stat','TP','TN'])

    en = time.time()
    print(str(en-st) + ' seconds')
    print('end HG/TP/TN')
    print('')
    print('begin filter')
    filt = time.time()
    output = output.sort_values(by='HG_stat', ascending=True)

    used_genes = []
    counter=0
    prev_stat = 0
    dropped=0
    for index, row in output.iterrows():
        #row[0] = gene1
        #row[1] = gene2
        #row[2] = gene3
        if counter == 1000:
            break
        #if row[0] != 'LY6D' or row[1] != 'CD3G_c':
        #    output.drop([index],inplace=True)
        #    dropped=dropped + 1
        #    print(dropped)
        #    continue
        #if row[-1] < .9:
        #    output.drop([index],inplace=True)
        #    continue
        if row[0]==row[1] or row[1]==row[2] or row[0]==row[2] or row[0]==row[3] or row[1]==row[3] or row[2]==row[3]:
            output.drop([index],inplace=True)
            continue
        if row[3] == prev_stat:
            output.drop([index],inplace=True)
            continue
        else:
            prev_stat = row[3]
        counter = counter+1

    endfilt = time.time()
    print(str(endfilt-filt) + ' seconds')
    print('end filter')
    return output
